# Rekall Memory Forensics
#
# Authors:
# Michael Cohen <scudette@users.sourceforge.net>
# Copyright (c) 2010, 2011, 2012 Michael Ligh <michael.ligh@mnin.org>
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

# pylint: disable=protected-access

"""References:
http://msdn.microsoft.com/en-us/magazine/ms809762.aspx
http://msdn.microsoft.com/en-us/magazine/cc301805.aspx
http://code.google.com/p/corkami/downloads/detail?name=pe-20110117.pdf
http://code.google.com/p/pefile/

Version information:
http://msdn.microsoft.com/en-us/library/windows/desktop/ff468916(v=vs.85).aspx
"""
import copy

from rekall import addrspace
from rekall import obj
from rekall import utils
from rekall.plugins.overlays import basic


class SentinalArray(obj.Array):
    """A sential terminated array."""

    def __iter__(self):
        """Break when the sentinal is reached."""
        for member in super(SentinalArray, self).__iter__():
            data = member.obj_vm.read(member.obj_offset, member.size())
            if data == "\x00" * member.size():
                break

            yield member


class RVAPointer(obj.Pointer):
    """A pointer through a relative virtual address."""
    image_base = 0

    def __init__(self, image_base=None, **kwargs):
        super(RVAPointer, self).__init__(**kwargs)
        self.image_base = image_base or self.obj_context.get("image_base", 0)

    def v(self, vm=None):
        rva_pointer = super(RVAPointer, self).v()
        if rva_pointer:
            rva_pointer += self.image_base

        return rva_pointer

class ResourcePointer(obj.Pointer):
    """A pointer relative to our resource section."""
    resource_base = 0

    def __init__(self, resource_base=None, **kwargs):
        super(ResourcePointer, self).__init__(**kwargs)
        # By default find the resource_base from the context.
        self.resource_base = (resource_base or
                              self.obj_context.get("resource_base"))

        if self.resource_base is None:
            for parent in self.parents:
                if isinstance(parent, _IMAGE_NT_HEADERS):
                    for section in parent.Sections:
                        if section.Name.startswith(".rsrc"):
                            self.resource_base = (section.VirtualAddress +
                                                  parent.OptionalHeader.ImageBase)
                            self.obj_context['resource_base'] = self.resource_base
                            break

    def v(self):
        # Only the first 31 bits are meaningful.
        resource_pointer = int(super(ResourcePointer, self).v()) & ((1 << 31) - 1)
        if resource_pointer:
            resource_pointer += self.resource_base

        return resource_pointer


pe_overlays = {
    "_IMAGE_OPTIONAL_HEADER": [ None, {
            'Subsystem' : [ None, ['Enumeration', {
                        'choices': {
                            0: 'IMAGE_SUBSYSTEM_UNKNOWN',
                            1: 'IMAGE_SUBSYSTEM_NATIVE',
                            2: 'IMAGE_SUBSYSTEM_WINDOWS_GUI',
                            3: 'IMAGE_SUBSYSTEM_WINDOWS_CUI',
                            5: 'IMAGE_SUBSYSTEM_OS2_CUI',
                            7: 'IMAGE_SUBSYSTEM_POSIX_CUI',
                            9: 'IMAGE_SUBSYSTEM_WINDOWS_CE_GUI',
                            10:'IMAGE_SUBSYSTEM_EFI_APPLICATION',
                            11:'IMAGE_SUBSYSTEM_EFI_BOOT_SERVICE_DRIVER',
                            12:'IMAGE_SUBSYSTEM_EFI_RUNTIME_DRIVER',
                            13:'IMAGE_SUBSYSTEM_EFI_ROM',
                            14:'IMAGE_SUBSYSTEM_XBOX'},
                        'target': 'unsigned short'}]],

            'DataDirectory' : [None, ['IndexedArray', {
                        'count': 16,
                        'index_table': {
                            'IMAGE_DIRECTORY_ENTRY_EXPORT':        0,
                            'IMAGE_DIRECTORY_ENTRY_IMPORT':        1,
                            'IMAGE_DIRECTORY_ENTRY_RESOURCE':      2,
                            'IMAGE_DIRECTORY_ENTRY_EXCEPTION':     3,
                            'IMAGE_DIRECTORY_ENTRY_SECURITY':      4,
                            'IMAGE_DIRECTORY_ENTRY_BASERELOC':     5,
                            'IMAGE_DIRECTORY_ENTRY_DEBUG':         6,
                            'IMAGE_DIRECTORY_ENTRY_COPYRIGHT':     7,
                            'IMAGE_DIRECTORY_ENTRY_GLOBALPTR':     8,
                            'IMAGE_DIRECTORY_ENTRY_TLS':           9,
                            'IMAGE_DIRECTORY_ENTRY_LOAD_CONFIG':   10,
                            'IMAGE_DIRECTORY_ENTRY_BOUND_IMPORT':  11,
                            'IMAGE_DIRECTORY_ENTRY_IAT':           12,
                            'IMAGE_DIRECTORY_ENTRY_DELAY_IMPORT':  13,
                            'IMAGE_DIRECTORY_ENTRY_COM_DESCRIPTOR':14,
                            'IMAGE_DIRECTORY_ENTRY_RESERVED':      15,
                            },
                        'target': '_IMAGE_DATA_DIRECTORY'}]],
            }],
    "_IMAGE_FILE_HEADER": [None, {
            'Machine' : [ 0x0, ['Enumeration', {
                        'choices': {
                            0x0000: 'IMAGE_FILE_MACHINE_UNKNOWN',
                            0x01d3: 'IMAGE_FILE_MACHINE_AM33',
                            0x8664: 'IMAGE_FILE_MACHINE_AMD64',
                            0x01c0: 'IMAGE_FILE_MACHINE_ARM',
                            0x0ebc: 'IMAGE_FILE_MACHINE_EBC',
                            0x014c: 'IMAGE_FILE_MACHINE_I386',
                            0x0200: 'IMAGE_FILE_MACHINE_IA64',
                            0x9041: 'IMAGE_FILE_MACHINE_MR32',
                            0x0266: 'IMAGE_FILE_MACHINE_MIPS16',
                            0x0366: 'IMAGE_FILE_MACHINE_MIPSFPU',
                            0x0466: 'IMAGE_FILE_MACHINE_MIPSFPU16',
                            0x01f0: 'IMAGE_FILE_MACHINE_POWERPC',
                            0x01f1: 'IMAGE_FILE_MACHINE_POWERPCFP',
                            0x0166: 'IMAGE_FILE_MACHINE_R4000',
                            0x01a2: 'IMAGE_FILE_MACHINE_SH3',
                            0x01a3: 'IMAGE_FILE_MACHINE_SH3DSP',
                            0x01a6: 'IMAGE_FILE_MACHINE_SH4',
                            0x01a8: 'IMAGE_FILE_MACHINE_SH5',
                            0x01c2: 'IMAGE_FILE_MACHINE_THUMB',
                            0x0169: 'IMAGE_FILE_MACHINE_WCEMIPSV2'},
                        'target': 'unsigned short'
                        }]],

            'Characteristics' : [ 0x12, ['Flags', {
                        'maskmap': {
                            'IMAGE_FILE_RELOCS_STRIPPED': 0x0001,
                            'IMAGE_FILE_EXECUTABLE_IMAGE': 0x0002,
                            'IMAGE_FILE_LINE_NUMS_STRIPPED': 0x0004,
                            'IMAGE_FILE_LOCAL_SYMS_STRIPPED': 0x0008,
                            'IMAGE_FILE_AGGRESIVE_WS_TRIM': 0x0010,
                            'IMAGE_FILE_LARGE_ADDRESS_AWARE': 0x0020,
                            'IMAGE_FILE_16BIT_MACHINE': 0x0040,
                            'IMAGE_FILE_BYTES_REVERSED_LO': 0x0080,
                            'IMAGE_FILE_32BIT_MACHINE': 0x0100,
                            'IMAGE_FILE_DEBUG_STRIPPED': 0x0200,
                            'IMAGE_FILE_REMOVABLE_RUN_FROM_SWAP': 0x0400,
                            'IMAGE_FILE_NET_RUN_FROM_SWAP': 0x0800,
                            'IMAGE_FILE_SYSTEM': 0x1000,
                            'IMAGE_FILE_DLL': 0x2000,
                            'IMAGE_FILE_UP_SYSTEM_ONLY': 0x4000,
                            'IMAGE_FILE_BYTES_REVERSED_HI': 0x8000,
                            },
                        'target': 'unsigned short'}]],
            'TimeDateStamp' : [ 0x4, ['UnixTimeStamp', {}]],
            }],

    "_IMAGE_SECTION_HEADER": [None, {
            'Name' : [ 0x0, ['String', {'length': 8, 'term': None}]],
            'Characteristics' : [ 0x24, ['Flags', {
                        'maskmap': {
                            'IMAGE_SCN_CNT_CODE':                  0x00000020,
                            'IMAGE_SCN_CNT_INITIALIZED_DATA':      0x00000040,
                            'IMAGE_SCN_CNT_UNINITIALIZED_DATA':    0x00000080,
                            'IMAGE_SCN_LNK_OTHER':                 0x00000100,
                            'IMAGE_SCN_LNK_INFO':                  0x00000200,
                            'IMAGE_SCN_LNK_REMOVE':                0x00000800,
                            'IMAGE_SCN_LNK_COMDAT':                0x00001000,
                            'IMAGE_SCN_MEM_FARDATA':               0x00008000,
                            'IMAGE_SCN_MEM_PURGEABLE':             0x00020000,
                            'IMAGE_SCN_MEM_16BIT':                 0x00020000,
                            'IMAGE_SCN_MEM_LOCKED':                0x00040000,
                            'IMAGE_SCN_MEM_PRELOAD':               0x00080000,
                            'IMAGE_SCN_ALIGN_1BYTES':              0x00100000,
                            'IMAGE_SCN_ALIGN_2BYTES':              0x00200000,
                            'IMAGE_SCN_ALIGN_4BYTES':              0x00300000,
                            'IMAGE_SCN_ALIGN_8BYTES':              0x00400000,
                            'IMAGE_SCN_ALIGN_16BYTES':             0x00500000,
                            'IMAGE_SCN_ALIGN_32BYTES':             0x00600000,
                            'IMAGE_SCN_ALIGN_64BYTES':             0x00700000,
                            'IMAGE_SCN_ALIGN_128BYTES':            0x00800000,
                            'IMAGE_SCN_ALIGN_256BYTES':            0x00900000,
                            'IMAGE_SCN_ALIGN_512BYTES':            0x00A00000,
                            'IMAGE_SCN_ALIGN_1024BYTES':           0x00B00000,
                            'IMAGE_SCN_ALIGN_2048BYTES':           0x00C00000,
                            'IMAGE_SCN_ALIGN_4096BYTES':           0x00D00000,
                            'IMAGE_SCN_ALIGN_8192BYTES':           0x00E00000,
                            'IMAGE_SCN_ALIGN_MASK':                0x00F00000,
                            'IMAGE_SCN_LNK_NRELOC_OVFL':           0x01000000,
                            'IMAGE_SCN_MEM_DISCARDABLE':           0x02000000,
                            'IMAGE_SCN_MEM_NOT_CACHED':            0x04000000,
                            'IMAGE_SCN_MEM_NOT_PAGED':             0x08000000,
                            'IMAGE_SCN_MEM_SHARED':                0x10000000,
                            'IMAGE_SCN_MEM_EXECUTE':               0x20000000,
                            'IMAGE_SCN_MEM_READ':                  0x40000000,
                            'IMAGE_SCN_MEM_WRITE':                 0x80000000L},
                        'target': 'unsigned long'}]],
            }],

    "_IMAGE_IMPORT_DESCRIPTOR": [None, {
            'Name': [ 0xC, ['RVAPointer', dict(target="String",
                                               target_args=dict(length=128))]],

            # This is an RVA pointer to an array of _IMAGE_THUNK_DATA structs.
            'FirstThunk': [ 0x10, ['RVAPointer', dict(target="ThunkArray")]],

            # This is a copy of the original IAT in memory.
            'OriginalFirstThunk': [ 0x0, ['RVAPointer', dict(target="ThunkArray")]],
            }],

    "_IMAGE_EXPORT_DIRECTORY": [None, {
            'Name': [ 0xC, ['RVAPointer', dict(target="String",
                                               target_args=dict(length=128))]],

            'AddressOfFunctions': [ None, ['RVAPointer', dict(
                        target="Array",
                        target_args=dict(target="RVAPointer",
                                         target_args=dict(target="Function"),
                                         count=lambda x: x.NumberOfFunctions)
                        )]],

            'AddressOfNames': [ None, ["RVAPointer", dict(
                        target="Array",
                        target_args=dict(
                            target="RVAPointer",
                            target_args=dict(target="String"),
                            count=lambda x: x.NumberOfNames,
                            )
                        )]],

            'AddressOfNameOrdinals': [ None, ['RVAPointer', dict(
                        target="Array",
                        target_args=dict(
                            target="unsigned short int",
                            count=lambda x: x.NumberOfFunctions)
                        )]],
            }],

    "_IMAGE_THUNK_DATA": [None, {
            'AddressOfData' : [ 0x0, ['RVAPointer', dict(target="_IMAGE_IMPORT_BY_NAME")]],
            }],

    "_IMAGE_THUNK_DATA64": [None, {
            'AddressOfData' : [ 0x0, ['RVAPointer', dict(target="_IMAGE_IMPORT_BY_NAME")]],
            }],

    "_IMAGE_NT_HEADERS": [None, {
            # This is a psuedo member to give access to the sections.
            "Sections": [
                # The sections start immediately after the OptionalHeader:
                lambda x: x.FileHeader.SizeOfOptionalHeader + x.OptionalHeader.obj_offset,

                # The sections are an array of _IMAGE_SECTION_HEADER structs.
                # The number of sections is found in the FileHeader
                ['Array', dict(target="_IMAGE_SECTION_HEADER",
                               count=lambda x: x.FileHeader.NumberOfSections)]],
            }],

    "_IMAGE_RESOURCE_DIRECTORY": [0x10, {
            "NumberOfNamedEntries": [0x0c, ['unsigned short int']],
            "NumberOfIdEntries": [0x0e, ['unsigned short int']],
            "Entries": [0x10, ["Array", dict(
                        target="_IMAGE_RESOURCE_DIRECTORY_ENTRY",
                        count=lambda x: x.NumberOfIdEntries + x.NumberOfNamedEntries)]],
            }],

    "_IMAGE_RESOURCE_DIRECTORY_ENTRY": [0x08, {
            "Name": [0x00, ['ResourcePointer', dict(target="PrefixedString")]],
            "Type": [0x00, ["Enumeration", dict(choices={
                            1:    'RT_CURSOR',
                            2:    'RT_BITMAP',
                            3:    'RT_ICON',
                            4:    'RT_MENU',
                            5:    'RT_DIALOG',
                            6:    'RT_STRING',
                            7:    'RT_FONTDIR',
                            8:    'RT_FONT',
                            9:    'RT_ACCELERATOR',
                            10:   'RT_RCDATA',
                            11:   'RT_MESSAGETABLE',
                            12:   'RT_GROUP_CURSOR',
                            14:   'RT_GROUP_ICON',
                            16:   'RT_VERSION',
                            17:   'RT_DLGINCLUDE',
                            19:   'RT_PLUGPLAY',
                            20:   'RT_VXD',
                            21:   'RT_ANICURSOR',
                            22:   'RT_ANIICON',
                            23:   'RT_HTML',
                            24:   'RT_MANIFEST'})]],

            # This is true when we need to use the Name field.
            "NameIsString": [0x00, ['BitField', dict(start_bit=31, end_bit=32)]],
            "OffsetToDataInt": [0x04, ['unsigned int']],
            "OffsetToData": [0x04, ['ResourcePointer', dict(target="_IMAGE_RESOURCE_DATA_ENTRY")]],
            "Entry": [0x04, ['ResourcePointer', dict(target="_IMAGE_RESOURCE_DIRECTORY")]],

            # If this is set the child is another _IMAGE_RESOURCE_DIRECTORY_ENTRY
            "ChildIsEntry": [0x04, ['BitField', dict(start_bit=31, end_bit=32)]],
            }],

    'PrefixedString' : [ 0x02, {
            'Length' : [ 0x0, ['unsigned short']],
            'Buffer' : [ 0x2, ['UnicodeString', dict(length=lambda x: x.Length * 2 + 1)]],
            } ],

    '_IMAGE_RESOURCE_DATA_ENTRY': [0x10, {
            'OffsetToData': [0x00, ['RVAPointer', dict(
                        target="String",
                        target_args=dict(length=lambda x: x.Size))]],
            'Size': [0x04, ['unsigned int']],
            'CodePage': [0x08, ['unsigned int']],
            }],
    }

# _IMAGE_OPTIONAL_HEADER64 is the same as _IMAGE_OPTIONAL_HEADER but offsets are
# different
pe_overlays["_IMAGE_OPTIONAL_HEADER64"] = copy.deepcopy(
    pe_overlays["_IMAGE_OPTIONAL_HEADER"])


def RoundUp(offset):
    """Round up the next word boundary."""
    if offset % 4:
        offset += 4 - offset % 4

    return offset


def AlignAfter(name):
    def get_offset(x):
        x = getattr(x, name)
        end_of_object = x.obj_offset + x.size()

        return RoundUp(end_of_object)

    return get_offset


pe_vtypes = {
    '_IMAGE_EXPORT_DIRECTORY': [ 0x28, {
            'Base': [ 0x10, ['unsigned int']],
            'NumberOfFunctions': [ 0x14, ['unsigned int']],
            'NumberOfNames': [ 0x18, ['unsigned int']],
            'AddressOfFunctions': [ 0x1C, ['unsigned int']],
            'AddressOfNames': [ 0x20, ['unsigned int']],
            'AddressOfNameOrdinals': [ 0x24, ['unsigned int']],
            }],

    '_IMAGE_IMPORT_DESCRIPTOR': [ 0x14, {
            'TimeDateStamp': [ 0x4, ['UnixTimeStamp', {}]],
            'ForwarderChain': [ 0x8, ['unsigned int']],
            }],

    # This is really a union of members.
    '_IMAGE_THUNK_DATA' : [ 0x4, {
            # Fake member for testing if the highest bit is set
            'OrdinalBit' : [ 0x0, ['BitField', dict(start_bit = 31, end_bit = 32)]],
            'Function' : [ 0x0, ['pointer', ['void']]],
            'Ordinal' : [ 0x0, ['unsigned long']],
            'AddressOfData' : [ 0x0, ['unsigned long']],
            'ForwarderString' : [ 0x0, ['unsigned int']],
            }],

    '_IMAGE_IMPORT_BY_NAME' : [ 4, {
            'Hint' : [ 0x0, ['unsigned short']],
            'Name' : [ 0x2, ['String', dict(length = 128)]],
            }],

    '__unnamed_156e' : [ 0x4, {
            'PhysicalAddress' : [ 0x0, ['unsigned long']],
            'VirtualSize' : [ 0x0, ['unsigned long']],
            } ],

    '_IMAGE_SECTION_HEADER' : [ 0x28, {
            'Name' : [ 0x0, ['String', {'length': 8, 'term': None}]],
            'Misc' : [ 0x8, ['__unnamed_156e']],
            'VirtualAddress' : [ 0xc, ['unsigned long']],
            'SizeOfRawData' : [ 0x10, ['unsigned long']],
            'PointerToRawData' : [ 0x14, ['unsigned long']],
            'PointerToRelocations' : [ 0x18, ['unsigned long']],
            'PointerToLinenumbers' : [ 0x1c, ['unsigned long']],
            'NumberOfRelocations' : [ 0x20, ['unsigned short']],
            'NumberOfLinenumbers' : [ 0x22, ['unsigned short']],
            } ],

    '_IMAGE_DOS_HEADER' : [ 0x40, {
            'e_magic' : [ 0x0, ['unsigned short']],
            'e_cblp' : [ 0x2, ['unsigned short']],
            'e_cp' : [ 0x4, ['unsigned short']],
            'e_crlc' : [ 0x6, ['unsigned short']],
            'e_cparhdr' : [ 0x8, ['unsigned short']],
            'e_minalloc' : [ 0xa, ['unsigned short']],
            'e_maxalloc' : [ 0xc, ['unsigned short']],
            'e_ss' : [ 0xe, ['unsigned short']],
            'e_sp' : [ 0x10, ['unsigned short']],
            'e_csum' : [ 0x12, ['unsigned short']],
            'e_ip' : [ 0x14, ['unsigned short']],
            'e_cs' : [ 0x16, ['unsigned short']],
            'e_lfarlc' : [ 0x18, ['unsigned short']],
            'e_ovno' : [ 0x1a, ['unsigned short']],
            'e_res' : [ 0x1c, ['array', 4, ['unsigned short']]],
            'e_oemid' : [ 0x24, ['unsigned short']],
            'e_oeminfo' : [ 0x26, ['unsigned short']],
            'e_res2' : [ 0x28, ['array', 10, ['unsigned short']]],
            'e_lfanew' : [ 0x3c, ['long']],
            } ],

    '_IMAGE_NT_HEADERS' : [ 0xf8, {
            'Signature' : [ 0x0, ['unsigned long']],
            'FileHeader' : [ 0x4, ['_IMAGE_FILE_HEADER']],
            'OptionalHeader' : [ 0x18, ['_IMAGE_OPTIONAL_HEADER']],
            } ],

    '_IMAGE_NT_HEADERS64' : [ 0x108, {
            'Signature' : [ 0x0, ['unsigned long']],
            'FileHeader' : [ 0x4, ['_IMAGE_FILE_HEADER']],
            'OptionalHeader' : [ 0x18, ['_IMAGE_OPTIONAL_HEADER64']],
            } ],

    '_IMAGE_OPTIONAL_HEADER64' : [ 0xf0, {
            'Magic' : [ 0x0, ['unsigned short']],
            'MajorLinkerVersion' : [ 0x2, ['unsigned char']],
            'MinorLinkerVersion' : [ 0x3, ['unsigned char']],
            'SizeOfCode' : [ 0x4, ['unsigned long']],
            'SizeOfInitializedData' : [ 0x8, ['unsigned long']],
            'SizeOfUninitializedData' : [ 0xc, ['unsigned long']],
            'AddressOfEntryPoint' : [ 0x10, ['unsigned long']],
            'BaseOfCode' : [ 0x14, ['unsigned long']],
            'ImageBase' : [ 0x18, ['unsigned long long']],
            'SectionAlignment' : [ 0x20, ['unsigned long']],
            'FileAlignment' : [ 0x24, ['unsigned long']],
            'MajorOperatingSystemVersion' : [ 0x28, ['unsigned short']],
            'MinorOperatingSystemVersion' : [ 0x2a, ['unsigned short']],
            'MajorImageVersion' : [ 0x2c, ['unsigned short']],
            'MinorImageVersion' : [ 0x2e, ['unsigned short']],
            'MajorSubsystemVersion' : [ 0x30, ['unsigned short']],
            'MinorSubsystemVersion' : [ 0x32, ['unsigned short']],
            'Win32VersionValue' : [ 0x34, ['unsigned long']],
            'SizeOfImage' : [ 0x38, ['unsigned long']],
            'SizeOfHeaders' : [ 0x3c, ['unsigned long']],
            'CheckSum' : [ 0x40, ['unsigned long']],
            'Subsystem' : [ 0x44, ['unsigned short']],
            'DllCharacteristics' : [ 0x46, ['unsigned short']],
            'SizeOfStackReserve' : [ 0x48, ['unsigned long long']],
            'SizeOfStackCommit' : [ 0x50, ['unsigned long long']],
            'SizeOfHeapReserve' : [ 0x58, ['unsigned long long']],
            'SizeOfHeapCommit' : [ 0x60, ['unsigned long long']],
            'LoaderFlags' : [ 0x68, ['unsigned long']],
            'NumberOfRvaAndSizes' : [ 0x6c, ['unsigned long']],
            'DataDirectory' : [ 0x70, ['array', 16, ['_IMAGE_DATA_DIRECTORY']]],
            } ],

    '_IMAGE_OPTIONAL_HEADER' : [ 0xe0, {
            'Magic' : [ 0x0, ['unsigned short']],
            'MajorLinkerVersion' : [ 0x2, ['unsigned char']],
            'MinorLinkerVersion' : [ 0x3, ['unsigned char']],
            'SizeOfCode' : [ 0x4, ['unsigned long']],
            'SizeOfInitializedData' : [ 0x8, ['unsigned long']],
            'SizeOfUninitializedData' : [ 0xc, ['unsigned long']],
            'AddressOfEntryPoint' : [ 0x10, ['unsigned long']],
            'BaseOfCode' : [ 0x14, ['unsigned long']],
            'BaseOfData' : [ 0x18, ['unsigned long']],
            'ImageBase' : [ 0x1c, ['unsigned long']],
            'SectionAlignment' : [ 0x20, ['unsigned long']],
            'FileAlignment' : [ 0x24, ['unsigned long']],
            'MajorOperatingSystemVersion' : [ 0x28, ['unsigned short']],
            'MinorOperatingSystemVersion' : [ 0x2a, ['unsigned short']],
            'MajorImageVersion' : [ 0x2c, ['unsigned short']],
            'MinorImageVersion' : [ 0x2e, ['unsigned short']],
            'MajorSubsystemVersion' : [ 0x30, ['unsigned short']],
            'MinorSubsystemVersion' : [ 0x32, ['unsigned short']],
            'Win32VersionValue' : [ 0x34, ['unsigned long']],
            'SizeOfImage' : [ 0x38, ['unsigned long']],
            'SizeOfHeaders' : [ 0x3c, ['unsigned long']],
            'CheckSum' : [ 0x40, ['unsigned long']],
            'Subsystem' : [ 0x44, ['unsigned long']],
            'DllCharacteristics' : [ 0x46, ['unsigned short']],
            'SizeOfStackReserve' : [ 0x48, ['unsigned long']],
            'SizeOfStackCommit' : [ 0x4c, ['unsigned long']],
            'SizeOfHeapReserve' : [ 0x50, ['unsigned long']],
            'SizeOfHeapCommit' : [ 0x54, ['unsigned long']],
            'LoaderFlags' : [ 0x58, ['unsigned long']],
            'NumberOfRvaAndSizes' : [ 0x5c, ['unsigned long']],
            'DataDirectory' : [ 0x60, ['unsigned long']],
            } ],

    '_IMAGE_FILE_HEADER' : [ 0x14, {
            'NumberOfSections' : [ 0x2, ['unsigned short']],
            'TimeDateStamp' : [ 0x4, ['UnixTimeStamp', {}]],
            'PointerToSymbolTable' : [ 0x8, ['unsigned long']],
            'NumberOfSymbols' : [ 0xc, ['unsigned long']],
            'SizeOfOptionalHeader' : [ 0x10, ['unsigned short']],
            } ],

    '_IMAGE_DATA_DIRECTORY' : [ 0x8, {
            'VirtualAddress' : [ 0x0, ['RVAPointer', dict(target='unsigned int')]],
            'Size' : [ 0x4, ['unsigned long']],
            } ],

    '_IMAGE_THUNK_DATA64' : [ 0x8, {
            # Fake member for testing if the highest bit is set
            'OrdinalBit' : [ 0x0, ['BitField', dict(start_bit = 63, end_bit = 64)]],
            'Function' : [ 0x0, ['pointer64', ['void']]],
            'Ordinal' : [ 0x0, ['unsigned long long']],
            'AddressOfData' : [ 0x0, ['unsigned long long']],
            'ForwarderString' : [ 0x0, ['unsigned long long']],
            }],

    # Note this is a problematic structure due to the alignment
    # requirements. Its not too much of a problem for the Rekall Memory Forensics object
    # system though :-)

    # http://msdn.microsoft.com/en-us/library/windows/desktop/ms647001(v=vs.85).aspx
    'VS_VERSIONINFO': [0x06, {
            "Length": [0x00, ['unsigned short int']],
            "ValueLength": [0x02, ['unsigned short int']],
            "Type": [0x04, ["Enumeration", dict(
                        choices={
                            0: "Binary",
                            1: "Text",
                            },
                        target='unsigned short int')]],

            # Must be VS_VERSION_INFO\x00 in utf16
            "szKey": [0x06, ["UnicodeString", dict(length=32)]],

            # The member is 32bit aligned after the szKey member.
            "Value": [AlignAfter("szKey"), ["VS_FIXEDFILEINFO"]],

            # This member is also aligned after the Value member.
            "Children": [AlignAfter("Value"), ['ListArray', dict(
                        target="StringFileInfo",
                        maximum_offset=lambda x: x.Length + x.obj_offset)]],
            }],

    'VS_FIXEDFILEINFO': [0x34, {
            "Signature": [0x00, ['unsigned int']],
            "StructVersion": [0x04, ['unsigned int']],
            "FileVersionMS": [0x08, ['unsigned int']],
            "FileVersionLS": [0x0c, ['unsigned int']],
            "ProductVersionMS": [0x10, ['unsigned int']],
            "ProductVersionLS": [0x14, ['unsigned int']],
            "FileFlagsMask": [0x18, ['unsigned int']],
            "FileFlags": [0x1c, ['unsigned int']],
            "FileOS": [0x20, ["Flags", dict(
                        maskmap={
                            "VOS_DOS": 0x00010000,
                            "VOS_NT": 0x00040000,
                            "VOS__WINDOWS16": 0x00000001,
                            "VOS__WINDOWS32": 0x00000004,
                            },
                        target='unsigned int')]],
            "FileType": [0x24, ['Enumeration', dict(
                        choices={
                            1: "VFT_APP (Application)",
                            2: "VFT_DLL (DLL)",
                            3: "VFT_DRV (Driver)",
                            4: "VFT_FORT (Font)",
                            5: "VFT_VXD",
                            7: "VFT_STATIC_LIB",
                            },
                        target='unsigned int')]],
            "FileSubtype": [0x28, ['unsigned int']],
            "FileDateMS": [0x2c, ['unsigned int']],
            "FileDateLS": [0x30, ['unsigned int']],
            "FileDate": [0x2c, ['WinTimeStamp', {}]],
            }],

    # The size of this is given by the Length member.
    "StringFileInfo": [lambda x: RoundUp(x.Length), {
            "Length": [0x00, ['unsigned short int']],
            "ValueLength": [0x02, ['unsigned short int']],
            "Type": [0x04, ['unsigned short int']],

            # Must be "StringFileInfo"
            "Key": [0x06, ['UnicodeString', dict(length=28)]],

            "Children": [AlignAfter("Key"), ['ListArray', dict(
                        target='StringTable',
                        maximum_offset=lambda x: x.Length + x.obj_offset)]],
            }],

    # The size of this is given by the Length member.
    "VarFileInfo": [lambda x: RoundUp(x.Length), {
            "Length": [0x00, ['unsigned short int']],
            "ValueLength": [0x02, ['unsigned short int']],
            "Type": [0x04, ['unsigned short int']],

            # Must be "VarFileInfo"
            "Key": [0x06, ['UnicodeString', dict(length=24)]],

            "Children": [AlignAfter("Key"), ['ListArray', dict(
                        target='Var',
                        maximum_offset=lambda x: x.Length + x.obj_offset)]],
            }],

    # Round up the size of the struct to word alignment.
    "Var": [lambda x: RoundUp(x.Length), {
            "Length": [0x00, ['unsigned short int']],
            "ValueLength": [0x02, ['unsigned short int']],
            "Type": [0x04, ['unsigned short int']],

            # This is exactly Translation
            "Key": [0x06, ['UnicodeString', dict(length=24)]],

            "Value": [AlignAfter("Key"), ['String', dict(
                        length=lambda x: x.ValueLength, term=None)]],
            }],

    "StringTable": [lambda x: RoundUp(x.Length), {
            "Length": [0x00, ['unsigned short int']],
            "ValueLength": [0x02, ['unsigned short int']],
            "Type": [0x04, ['unsigned short int']],

            # In MSDN this is called szKey.
            "LangID": [0x06, ['UnicodeString', dict(length=16, term=None)]],

            "Children": [AlignAfter("LangID"), ['ListArray', dict(
                        target='ResourceString',
                        maximum_offset=lambda x: x.Length + x.obj_offset)]],
            }],

    # Round up the size of the struct to word alignment.
    "ResourceString": [lambda x: RoundUp(x.Length), {
            "Length": [0x00, ['unsigned short int']],
            "ValueLength": [0x02, ['unsigned short int']],
            "Type": [0x04, ['unsigned short int']],

            # This is a null terminated unicode string representing the key.
            "Key": [0x06, ['UnicodeString', dict(length=1024)]],

            "Value": [AlignAfter("Key"), ['UnicodeString', dict(
                        length=lambda x: x.ValueLength * 2)]],
            }],
    }


class _IMAGE_EXPORT_DIRECTORY(obj.Struct):
    """Class for PE export directory"""

    def valid(self, nt_header):
        """
        Check the sanity of export table fields.

        The RVAs cannot be larger than the module size. The function
        and name counts cannot be larger than 32K.
        """
        try:
            return (self.AddressOfFunctions < nt_header.OptionalHeader.SizeOfImage and
                    self.AddressOfNameOrdinals < nt_header.OptionalHeader.SizeOfImage and
                    self.AddressOfNames < nt_header.OptionalHeader.SizeOfImage and
                    self.NumberOfFunctions < 0x7FFF and
                    self.NumberOfNames < 0x7FFF)
        except obj.InvalidOffsetError:
            return False

    def _name(self, name_rva):
        """
        Return a String object for the function name.

        Names are truncated at 128 characters although its possible
        they may be longer. Thus, infrequently a function name will
        be missing some data. However, that's better than hard-coding
        a larger value which frequently causes us to cross page
        boundaries and return a NoneObject anyway.
        """
        return self.obj_profile.Object("String",
                                       offset = self.obj_parent.DllBase.v() + name_rva,
                                       vm = self.obj_vm, length = 128)

    def _exported_functions(self):
        """
        Generator for exported functions.

        @return: tuple (Ordinal, FunctionRVA, Name)

        Ordinal is an integer and should never be None. If the function
        is forwarded, FunctionRVA is None. Otherwise, FunctionRVA is an
        RVA to the function's code (relative to module base). Name is a
        String containing the exported function's name. If the Name is
        paged, it will be None. If the function is forwarded, Name is the
        forwarded function name including the DLL (ntdll.EtwLogTraceEvent).
        """

        mod_base = self.obj_parent.DllBase.v()
        exp_dir = self.obj_parent.export_dir()

        # PE files with a large number of functions will have arrays
        # that spans multiple pages. Thus the first entries may be valid,
        # last entries may be valid, but middle entries may be invalid
        # (paged). In the various checks below, we test for None (paged)
        # and zero (non-paged but invalid RVA).

        # Array of RVAs to function code
        address_of_functions = self.obj_profile.Object(
            'Array', offset = mod_base + self.AddressOfFunctions,
            target = 'unsigned int', count = self.NumberOfFunctions,
            vm = self.obj_vm)

        # Array of RVAs to function names
        address_of_names = self.obj_profile.Object(
            'Array', offset = mod_base + self.AddressOfNames,
            target = 'unsigned int', count = self.NumberOfNames,
            vm = self.obj_vm)

        # Array of RVAs to function ordinals
        address_of_name_ordinals = self.obj_profile.Object(
            'Array', offset = mod_base + self.AddressOfNameOrdinals,
            target = 'unsigned short', count = self.NumberOfNames,
            vm = self.obj_vm)

        # When functions are exported by Name, it will increase
        # NumberOfNames by 1 and NumberOfFunctions by 1. When
        # functions are exported by Ordinal, only the NumberOfFunctions
        # will increase. First we enum functions exported by Name
        # and track their corresponding Ordinals, so that when we enum
        # functions exported by Ordinal only, we don't duplicate.

        seen_ordinals = []

        # Handle functions exported by name *and* ordinal
        for i in range(self.NumberOfNames):

            name_rva = address_of_names[i]
            ordinal = address_of_name_ordinals[i]

            if name_rva in (0, None):
                continue

            # Check the sanity of ordinal values before using it as an index
            if ordinal == None or ordinal >= self.NumberOfFunctions:
                continue

            func_rva = address_of_functions[ordinal]

            if func_rva in (0, None):
                continue

            # Handle forwarded exports. If the function's RVA is inside the exports
            # section (as given by the VirtualAddress and Size fields in the
            # DataDirectory), the symbol is forwarded. Return the name of the
            # forwarded function and None as the function address.

            if (func_rva >= exp_dir.VirtualAddress and
                    func_rva < exp_dir.VirtualAddress + exp_dir.Size):
                n = self._name(func_rva)
                f = obj.NoneObject("This function is forwarded")
            else:
                n = self._name(name_rva)
                f = func_rva

            # Add the ordinal base and save it
            ordinal += self.Base
            seen_ordinals.append(ordinal)

            yield ordinal, f, n

        # Handle functions exported by ordinal only
        for i in range(self.NumberOfFunctions):

            ordinal = self.Base + i

            # Skip functions already enumberated above
            if ordinal not in seen_ordinals:

                func_rva = address_of_functions[i]

                if func_rva in (0, None):
                    continue

                seen_ordinals.append(ordinal)

                # There is no name RVA
                yield ordinal, func_rva, obj.NoneObject("Name RVA not accessible")


class _IMAGE_IMPORT_DESCRIPTOR(obj.Struct):
    """Handles IID entries for imported functions"""

    def valid(self, nt_header):
        """Check the validity of some fields"""
        try:
            return (self.OriginalFirstThunk != 0 and
                    self.OriginalFirstThunk < nt_header.OptionalHeader.SizeOfImage and
                    self.FirstThunk != 0 and
                    self.FirstThunk < nt_header.OptionalHeader.SizeOfImage and
                    self.Name < nt_header.OptionalHeader.SizeOfImage)
        except obj.InvalidOffsetError:
            return False

    def _name(self, name_rva):
        """Return a String object for the name at the given RVA"""

        return self.obj_profile.Object(
            "String", offset = self.obj_parent.DllBase.v() + name_rva,
            vm = self.obj_vm, length = 128)

    def dll_name(self):
        """Returns the name of the DLL for this IID"""
        return self._name(self.Name)

    def _imported_functions(self):
        """
        Generator for imported functions.

        @return: tuple (Ordinal, FunctionVA, Name)

        If the function is imported by ordinal, then Ordinal is the
        ordinal value and Name is None.

        If the function is imported by name, then Ordinal is the
        hint and Name is the imported function name (or None if its
        paged).

        FunctionVA is the virtual address of the imported function,
        as applied to the IAT by the Windows loader. If the FirstThunk
        is paged, then FunctionVA will be None.
        """

        i = 0
        while 1:
            thunk = self.obj_profile.Object(
                '_IMAGE_THUNK_DATA',
                offset = self.obj_parent.DllBase.v() + self.OriginalFirstThunk +
                i * self.obj_profile.get_obj_size('_IMAGE_THUNK_DATA'),
                vm = self.obj_vm)

            # We've reached the end when the element is zero
            if thunk == None or thunk.AddressOfData == 0:
                break

            o = obj.NoneObject("Ordinal not accessible?")
            n = obj.NoneObject("Imported by ordinal?")
            f = obj.NoneObject("FirstThunk not accessible")

            # If the highest bit (32 for x86 and 64 for x64) is set, the function is
            # imported by ordinal and the lowest 16-bits contain the ordinal value.
            # Otherwise, the lowest bits (0-31 for x86 and 0-63 for x64) contain an
            # RVA to an _IMAGE_IMPORT_BY_NAME struct.
            if thunk.OrdinalBit == 1:
                o = thunk.Ordinal & 0xFFFF
            else:
                iibn = self.obj_profile.Object("_IMAGE_IMPORT_BY_NAME",
                                               offset = self.obj_parent.DllBase.v() +
                                               thunk.AddressOfData,
                                               vm = self.obj_vm)
                o = iibn.Hint
                n = iibn.Name

            # See if the import is bound (i.e. resolved)
            first_thunk = self.obj_profile.Object(
                '_IMAGE_THUNK_DATA',
                offset = self.obj_parent.DllBase.v() + self.FirstThunk +
                i * self.obj_profile.get_obj_size('_IMAGE_THUNK_DATA'),
                vm = self.obj_vm)

            if first_thunk:
                f = first_thunk.Function.v()

            yield o, f, n
            i += 1

    def is_list_end(self):
        """Returns True if we've reached the list end"""
        data = self.obj_vm.read(
                        self.obj_offset,
                        self.obj_profile.get_obj_size('_IMAGE_IMPORT_DESCRIPTOR')
                        )
        return data.count(chr(0)) == len(data)


class _LDR_DATA_TABLE_ENTRY(obj.Struct):
    """
    Class for PE file / modules

    If these classes are instantiated by _EPROCESS.list_*_modules()
    then its guaranteed to be in the process address space.
    """
    _pe = None

    @property
    def PE(self):
        if self._pe is None:
            self._pe = PE(address_space=self.obj_vm, image_base=self.DllBase)

        return self._pe

    @property
    def NTHeader(self):
        """Return the _IMAGE_NT_HEADERS object"""

        dos_header = self.obj_profile.Object("_IMAGE_DOS_HEADER", offset = self.DllBase.v(),
                                             vm = self.obj_vm)

        return dos_header.NTHeader

    def _directory(self, dir_index):
        """Return the requested IMAGE_DATA_DIRECTORY"""
        data_dir = self.NTHeader.OptionalHeader.DataDirectory[dir_index]

        # Make sure the directory exists
        if data_dir.VirtualAddress == 0 or data_dir.Size == 0:
            raise ValueError('No export directory')

        # Make sure the directory VA and Size are sane
        if (data_dir.VirtualAddress + data_dir.Size >
            self.NTHeader.OptionalHeader.SizeOfImage):
            raise ValueError('Invalid Export directory')

        return data_dir

    def export_dir(self):
        """Return the IMAGE_DATA_DIRECTORY for exports"""
        return self._directory(0) # DIRECTORY_ENTRY_EXPORT

    def import_dir(self):
        """Return the IMAGE_DATA_DIRECTORY for imports"""
        return self._directory(1) # DIRECTORY_ENTRY_IMPORT

    def getprocaddress(self, func):
        """Return the RVA of func"""
        for _, f, n in self.exports():
            if str(n) == func:
                return f
        return None

    def imports(self):
        """
        Generator for the PE's imported functions.

        The _DIRECTORY_ENTRY_IMPORT.VirtualAddress points to an array
        of _IMAGE_IMPORT_DESCRIPTOR structures. The end is reached when
        the IID structure is all zeros.
        """

        try:
            data_dir = self.import_dir()
        except ValueError:
            return

        i = 0

        desc_size = self.obj_profile.get_obj_size('_IMAGE_IMPORT_DESCRIPTOR')

        while 1:
            desc = self.obj_profile.Object(
                '_IMAGE_IMPORT_DESCRIPTOR', vm = self.obj_vm,
                offset = self.DllBase.v() + data_dir.VirtualAddress + (i * desc_size),
                parent = self)

            # Stop if the IID is paged or all zeros
            if desc == None or desc.is_list_end():
                break

            # Stop if the IID contains invalid fields
            if not desc.valid(self.NTHeader):
                break

            dll_name = desc.dll_name()

            for o, f, n in desc._imported_functions():
                yield dll_name, o, f, n

            i += 1

    def exports(self):
        """Generator for the PE's exported functions"""

        try:
            data_dir = self.export_dir()
        except ValueError, why:
            raise StopIteration(why)

        expdir = self.obj_profile.Object(
            '_IMAGE_EXPORT_DIRECTORY', offset = self.DllBase.v() + data_dir.VirtualAddress,
            vm = self.obj_vm, parent = self)

        if expdir.valid(self.NTHeader):
            # Ordinal, Function RVA, and Name Object
            for o, f, n in expdir._exported_functions():
                yield o, f, n


class _IMAGE_DOS_HEADER(obj.Struct):
    """DOS header"""

    #Put checks in constructor.

    @property
    def NTHeader(self):
        """Get the NT header"""

        if self.e_magic != 0x5a4d:
            return obj.NoneObject('e_magic {0:04X} is not a valid DOS signature.'.format(
                    self.e_magic))

        nt_header = self.obj_profile._IMAGE_NT_HEADERS(
            offset=self.e_lfanew + self.obj_offset,
            vm=self.obj_vm, context=self.obj_context)

        if nt_header.Signature != 0x4550:
            return obj.NoneObject('NT header signature {0:04X} is not a valid'.format(
                    nt_header.Signature))

        return nt_header


class _IMAGE_NT_HEADERS(obj.Struct):
    """PE header"""

    @property
    def OptionalHeader(self):
        optional_header = self.m("OptionalHeader")
        if optional_header.Magic == 0x20b:
            return optional_header.cast("_IMAGE_OPTIONAL_HEADER64")

        return optional_header


class _IMAGE_SECTION_HEADER(obj.Struct):
    """PE section"""

    def sanity_check_section(self):
        """Sanity checks address boundaries"""
        # Note: all addresses here are RVAs
        image_size = self.obj_parent.OptionalHeader.SizeOfImage
        if self.VirtualAddress > image_size:
            raise ValueError('VirtualAddress {0:08x} is past the end of '
                             'image.'.format(self.VirtualAddress))

        if self.Misc.VirtualSize > image_size:
            raise ValueError('VirtualSize {0:08x} is larger than image '
                             'size.'.format(self.Misc.VirtualSize))

        if self.SizeOfRawData > image_size:
            raise ValueError('SizeOfRawData {0:08x} is larger than image '
                             'size.'.format(self.SizeOfRawData))


class _IMAGE_DATA_DIRECTORY(obj.Struct):
    """A data directory."""

    def dereference(self):
        """Automatically resolve the data directory according to our name."""
        result = self.m("VirtualAddress")

        if self.obj_name == "IMAGE_DIRECTORY_ENTRY_IMPORT":
            return result.dereference_as(
                target="SentinalArray", target_args=dict(
                    target="_IMAGE_IMPORT_DESCRIPTOR"
                    )
                )

        elif self.obj_name == "IMAGE_DIRECTORY_ENTRY_EXPORT":
            return result.dereference_as("_IMAGE_EXPORT_DIRECTORY")

        elif self.obj_name == "IMAGE_DIRECTORY_ENTRY_RESOURCE":
            return result.dereference_as("_IMAGE_RESOURCE_DIRECTORY")

        return result.dereference()


class _IMAGE_RESOURCE_DIRECTORY(obj.Struct):
    """Represents a node in the resource tree."""

    def __iter__(self):
        for entry in self.Entries:
            yield entry

    def Open(self, node_name):
        """Opens a specific node child."""
        for entry in self.Entries:
            if entry.Name == node_name or entry.Type == node_name:
                return entry.Entry

        return obj.NoneObject("node %s not found" % node_name)

    def Traverse(self):
        """A generator for _IMAGE_RESOURCE_DATA_ENTRY under this node."""
        for entry in self:
            if entry.ChildIsEntry:
                for subentry in entry.Entry.Traverse():
                    yield subentry
            else:
                yield entry.OffsetToData.dereference()


class _IMAGE_RESOURCE_DIRECTORY_ENTRY(obj.Struct):

    @property
    def Name(self):
        if self.NameIsString:
            return utils.SmartUnicode(self.m("Name").Buffer)
        else:
            return utils.SmartUnicode(self.Type)

    @property
    def Entry(self):
        if self.ChildIsEntry:
            return self.m("Entry").dereference()
        else:
            return self.m("OffsetToData")


class ThunkArray(SentinalArray):
    """A sential terminated array of thunks."""

    def __init__(self, parent=None, **kwargs):
        target="_IMAGE_THUNK_DATA"

        # Are we in a 64 bit file?
        for x in parent.parents:
            if x.obj_name.endswith("64"):
                target += "64"
                break

        super(ThunkArray, self).__init__(target=target, parent=parent,
                                         **kwargs)

class VS_VERSIONINFO(obj.Struct):

    @property
    def Children(self):
        """The child is either a StringFileInfo or VarFileInfo depending on the key."""
        for child in self.m("Children"):
            if child.Key.startswith("VarFileInfo"):
                yield child.cast("VarFileInfo")
            elif child.Key.startswith("StringFileInfo"):
                yield child
            else:
                break

    def Strings(self, obj=None):
        """Generates all the ResourceString structs by recursively traversing
        the Children tree.
        """
        if obj is None:
            obj = self

        for child in obj.Children:
            try:
                for subchild in self.Strings(child):
                    yield subchild
            except AttributeError:
                yield child



class PE(object):
    """A convenience object to access various things in a PE file."""

    def __init__(self, address_space=None, image_base=0, filename=None):
        """Constructor.

        Args:
          address_space: An address space to examine.

          image_base: The address of the dos header.

          filename: If a filename is provided we open the file as a PE File. In
            this case, image_base and address_space are ignored.
        """
        self.profile = PEProfile()

        if filename is None:
            self.vm = address_space
            self.image_base = image_base
        else:
            file_address_space = PEFileAddressSpace.classes[
                'FileAddressSpace'](filename=filename)
            self.vm = PEFileAddressSpace(base=file_address_space)
            self.image_base = self.vm.image_base

        self.dos_header = self.profile.Object(
            "_IMAGE_DOS_HEADER", vm=self.vm, offset=self.image_base,
            context=dict(image_base=self.image_base))

        self.nt_header = self.dos_header.NTHeader

    def ImportDirectory(self):
        """A generator over the import directory.

        Note that this iterates over the OriginalFirstThunk which still remains
        from the on-disk executable. The IAT is constructed by the linker at
        load time, and is stored in FirstThunk in memory. Hence the IAT() method
        is going to return code objects while this method simply returns names.

        Yields:
           a tuple of (dll, function_name)
        """
        import_directory = self.nt_header.OptionalHeader.DataDirectory[
            'IMAGE_DIRECTORY_ENTRY_IMPORT'].dereference()

        for directory in import_directory:
            dll = directory.Name.dereference()
            for thunk in directory.OriginalFirstThunk.dereference():
                function_name = thunk.AddressOfData.Name

                yield dll, function_name, thunk.AddressOfData.Hint

    def IAT(self):
        """A generator over the IAT.

        Note that this iterates over the FirstThunk imports. In memory, these
        contain the IAT which has been resolved by the loader.

        Yields:
          a tuple of (dll, function_name)
        """
        import_directory = self.nt_header.OptionalHeader.DataDirectory[
            'IMAGE_DIRECTORY_ENTRY_IMPORT'].dereference()

        for directory in import_directory:
            dll = directory.Name.dereference()
            for thunk in directory.FirstThunk.dereference():
                function = thunk.Function

                yield dll, function, thunk.Ordinal

    def ExportDirectory(self):
        """A generator over the export directory."""
        export_directory = self.nt_header.OptionalHeader.DataDirectory[
            'IMAGE_DIRECTORY_ENTRY_EXPORT'].dereference()

        dll = export_directory.Name.dereference()
        function_table = export_directory.AddressOfFunctions.dereference()
        name_table = export_directory.AddressOfNames.dereference()
        ordinal_table = export_directory.AddressOfNameOrdinals.dereference()

        seen_ordinals = set()

        # First do the names.
        for i, name in enumerate(name_table):
            ordinal = int(ordinal_table[i])
            seen_ordinals.add(ordinal)
            func = function_table[ordinal]
            func.obj_name = "%s:%s" % (dll, name.dereference())

            yield (dll, func, name.dereference(), ordinal)

        # Now the functions without names
        for i, func in enumerate(function_table):
            ordinal = export_directory.Base + i
            if ordinal in seen_ordinals:
                continue

            yield (dll, function_table[ordinal],
                   obj.NoneObject("Name not accessible"), ordinal)

    def GetProcAddress(self, name):
        """Scan the export table for a function of the given name.

        Similar to the GetProcAddress function.
        """
        for dll, function_pointer, func_name, ordinal in self.ExportDirectory():
            if func_name == name:
                return function_pointer.dereference()

    def VersionInformation(self):
        """A generator of key, value pairs."""
        resource_directory = self.nt_header.OptionalHeader.DataDirectory[
            'IMAGE_DIRECTORY_ENTRY_RESOURCE'].dereference()

        # Find all the versions and their strings
        for data in resource_directory.Open("RT_VERSION").Traverse():
            version_info = data.OffsetToData.dereference_as("VS_VERSIONINFO")
            for string in version_info.Strings():
                yield string.Key, string.Value

    def Sections(self):
        for section in self.nt_header.Sections:

            execution_flags = "%s%s%s" % (
                "x" if section.Characteristics.IMAGE_SCN_MEM_EXECUTE else "-",
                "r" if section.Characteristics.IMAGE_SCN_MEM_READ else "-",
                "w" if section.Characteristics.IMAGE_SCN_MEM_WRITE else "-")

            yield (execution_flags, section.Name, section.VirtualAddress,
                   section.SizeOfRawData)


# The following adds a profile to deal with PE files. Since PE files are not
# actually related to the kernel version, they get their own domain specific
# profile.
class PEFileImplementation(obj.ProfileModification):
    """An implementation of a parser for PE files."""

    @classmethod
    def Modify(cls, profile):
        profile.add_types(pe_vtypes)
        profile.add_classes({
                '_IMAGE_DOS_HEADER': _IMAGE_DOS_HEADER,
                '_IMAGE_NT_HEADERS': _IMAGE_NT_HEADERS,
                '_IMAGE_SECTION_HEADER': _IMAGE_SECTION_HEADER,
                '_IMAGE_EXPORT_DIRECTORY': _IMAGE_EXPORT_DIRECTORY,
                '_IMAGE_IMPORT_DESCRIPTOR': _IMAGE_IMPORT_DESCRIPTOR,
                '_LDR_DATA_TABLE_ENTRY': _LDR_DATA_TABLE_ENTRY,
                '_IMAGE_DATA_DIRECTORY': _IMAGE_DATA_DIRECTORY,
                "SentinalArray": SentinalArray,
                "ThunkArray": ThunkArray,
                "RVAPointer": RVAPointer,
                "ResourcePointer": ResourcePointer,
                "_IMAGE_RESOURCE_DIRECTORY": _IMAGE_RESOURCE_DIRECTORY,
                "_IMAGE_RESOURCE_DIRECTORY_ENTRY": _IMAGE_RESOURCE_DIRECTORY_ENTRY,
                "VS_VERSIONINFO": VS_VERSIONINFO,
                })
        profile.add_overlay(pe_overlays)

        return profile


class PEProfile(basic.Profile32Bits, basic.BasicWindowsClasses):
    """A profile for PE files."""

    def __init__(self, **kwargs):
        super(PEProfile, self).__init__(**kwargs)
        PEFileImplementation.Modify(self)


class PEFileAddressSpace(addrspace.BaseAddressSpace):
    """An address space which applies to PE files.

    This basically remaps sections in the PE file to the virtual address space.
    See http://code.google.com/p/corkami/downloads/detail?name=pe-20110117.pdf

    The PE file is divided into sections, each section is mapped into memory at
    a different place:

    File on Disk                 Memory Image
0-> ------------    image base-> ------------
     Header                      Header
    ------------                 ------------
     Section 1
    ------------                 ------------
     Section 2                    Section 1
    ------------                 ------------

                                 ------------
                                  Section 2
                                 ------------

    This address space expands the file from disk into the memory image view as
    shown. Since all internal pe RVA references are within the virtual space,
    this helps resolution.
    """
    def __init__(self, **kwargs):
        """We layer on top of the file address space."""
        super(PEFileAddressSpace, self).__init__(**kwargs)

        self.as_assert(self.base is not None, "Must layer on another AS.")
        self.as_assert(self.base.read(0, 2) == "MZ", "File does not have a PE signature.")
        self.profile = PEProfile()

        nt_header = self.profile.Object(
            "_IMAGE_DOS_HEADER", vm=self.base, offset=0).NTHeader

        self.image_base = int(nt_header.OptionalHeader.ImageBase)
        # Now map all the sections into a virtual address space.
        self.runs = []

        # The first run maps the file header over the base address
        self.runs.append(
            (self.image_base, nt_header.OptionalHeader.SizeOfHeaders, 0))

        for section in nt_header.Sections:
            virtual_address = section.VirtualAddress.v() + self.image_base
            self.runs.append(
                (virtual_address, section.SizeOfRawData.v(), section.PointerToRawData.v()))

        # TODO: The sections may overlap: What to do then?
        # Make sure that the sections are sorted.
        self.runs.sort()

        self.nt_header = self.profile.Object(
            "_IMAGE_DOS_HEADER", vm=self, offset=self.image_base,
            context=dict(image_base=0)).NTHeader

    def read(self, addr, length):
        # Not a particularly efficient algorithm, but probably fast enough since
        # usually there are not too many sections.
        for virtual_address, run_length, physical_address in self.runs:
            if addr >= virtual_address and addr <= virtual_address + run_length:
                offset = addr - virtual_address
                to_read = min(run_length - offset, length)

                return self.base.read(physical_address + offset, to_read)

        # Otherwise just null pad the results.
        return '\x00' * length



# Rekall Memory Forensics
# Copyright (C) 2012 Michael Cohen
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""This module guesses the current profile using various heuristics."""

__author__ = "Michael Cohen <scudette@gmail.com>"

# pylint: disable=protected-access

import logging

from rekall import addrspace

from rekall import plugin
from rekall import obj
from rekall import scan

from rekall.plugins.overlays import basic
from rekall.plugins.overlays.windows import windows


# The idea is to avoid importing all the plugins at once just to check a couple
# of offsets, so we just use an abbridged version of the full profiles.


class IdleScanner(scan.BaseScanner):
    def __init__(self, process_names=None, **kwargs):
        super(IdleScanner, self).__init__(**kwargs)
        self.checks = [('MultiStringFinderCheck', dict(
                    needles=process_names))]


class TestProfile32(basic.Profile32Bits, basic.BasicWindowsClasses):
    def __init__(self, **kwargs):
        super(TestProfile32, self).__init__(**kwargs)
        self.add_overlay(eprocess_vtypes)


class TestProfile64(basic.ProfileLLP64, basic.BasicWindowsClasses):
    def __init__(self, **kwargs):
        super(TestProfile64, self).__init__(**kwargs)
        self.add_overlay(eprocess_vtypes)


class GuessProfile(plugin.PhysicalASMixin, plugin.Command):
    """Guess the exact windows profile using heuristics."""

    __name = "guess_profile"

    DEFAULT_PROCESS_NAMES = ['System']

    def __init__(self, process_names=None, quick=True, **kwargs):
        """Guess the potential windows profiles which could be used.

        Args:
          process_names: A list of process names to search for - default System.

          quick: If set we just update the session profile and quit after the
            first result is found.
        """
        super(GuessProfile, self).__init__(**kwargs)
        self.process_names = process_names or self.DEFAULT_PROCESS_NAMES
        self.process_names = [
            x + "\x00" * (15 - len(x)) for x in self.process_names]

        self.quick = quick

    def guess_profile_from_kdbg(self, kdbg):
        """Guess the profile from the PsActiveProcessList pointer.

        Sometimes we can arrive at a dtb and even a kdbg without needing to know
        the profile at all (e.g. if thes were provided on the commandline or are
        known in advance from the image). In this case, it is not necessary to
        scan the image for the System process since we have a valid
        PsActiveProcessList. The following algorithm guesses the profile from
        the PsActiveProcessHead and it much faster.

        We assume a valid physical_address_space is already known here.

        Args:
          kdbg: A valid _KDDEBUGGER_DATA64 offset.
        """
        kdbg_profile = windows.KDDebuggerProfile()
        lookup_map = self._build_lookup_map()

        # First try to get a valid kernel address space. At this point we dont
        # even know if its a 32 bit or 64 bit profile so we try both address
        # spaces.
        for address_space_name in ["IA32PagedMemoryPae",
                                   "AMD64PagedMemory"]:
            address_space_cls = addrspace.BaseAddressSpace.GetPlugin(
                address_space_name)

            try:
                kernel_as = address_space_cls(
                    base=self.session.physical_address_space,
                    session=self.session)

                kdbg = kdbg_profile._KDDEBUGGER_DATA64(
                    offset=self.session.GetParameter("kdbg"),
                    vm=kernel_as)

                if not kdbg:
                    continue

                # Now try to see which profile fits the _EPROCESS best.
                for name, (test_profile, cls) in lookup_map.items():
                    active_process_list_offset = test_profile.get_obj_offset(
                        name, "ActiveProcessLinks")

                    # This is the first _EPROCESS linked from
                    # PsActiveProcessList.
                    eprocess = test_profile.Object(
                        name, vm=kernel_as,
                        offset=(kdbg.PsActiveProcessHead.Flink.v() -
                                active_process_list_offset))

                    if (eprocess.UniqueProcessId == 0 or
                        eprocess.UniqueProcessId > 10):
                        continue

                    dtb = eprocess.DirectoryTableBase.v()

                    # The first process should be the System process.
                    address_space = self.verify_profile(cls, eprocess, dtb)
                    if address_space:
                        yield name, kernel_as, eprocess

            except addrspace.ASAssertionError:
                pass

    def _build_lookup_map(self):
        test_profile32 = TestProfile32()
        test_profile64 = TestProfile64()

        # Precalculate the offsets to save time later.
        lookup_map = {}
        for name, cls in obj.Profile.classes.items():
            if cls.metadata("os") == "windows":
                if cls.metadata("memory_model") == "64bit":
                    test_profile = test_profile64
                else:
                    test_profile = test_profile32

                lookup_map[name] = (test_profile, cls)

        return lookup_map

    def guess_profile(self):
        """Guess using the Idle process technique."""
        # First ensure there is an address space
        if not self.session.physical_address_space:
            return

        scanner = IdleScanner(session=self.session,
                              process_names=self.process_names,
                              address_space=self.session.physical_address_space)

        lookup_map = self._build_lookup_map()
        logging.info("Searching for suitable System Process")
        for hit in scanner.scan():
            # Try every profile until one works.
            for name, (test_profile, cls) in lookup_map.items():
                try:
                    eprocess_offset = test_profile.get_obj_offset(
                        name, 'ImageFileName')
                except AttributeError:
                    logging.warning("Missing definition for profile %s", name)
                    continue

                eprocess = test_profile.Object(
                    name, vm=self.session.physical_address_space,
                    offset=hit - eprocess_offset)

                if (eprocess.UniqueProcessId == 0 or
                    eprocess.UniqueProcessId > 10):
                    continue

                dtb = eprocess.DirectoryTableBase.v()

                address_space = self.verify_profile(cls, eprocess, dtb)
                if address_space:
                    yield name, address_space, eprocess

    def verify_address_space(self, profile_cls, eprocess, address_space):
        """Check the eprocess for sanity."""
        # In windows the DTB must be page aligned, except for PAE images where
        # its aligned to a 0x20 size.
        if not profile_cls.metadata("pae") and address_space.dtb & 0xFFF != 0:
            return False

        if profile_cls.metadata("pae") and address_space.dtb & 0xF != 0:
            return False

        # Reflect through the address space at ourselves.
        list_head = eprocess.ThreadListHead.Flink
        if list_head == 0:
            return False

        me = list_head.dereference(vm=address_space).Blink.Flink
        if me.v() != list_head.v():
            return False

        return True

    def verify_profile(self, profile_cls, eprocess, dtb):
        """Verify potential profiles against the dtb hit."""
        # Make an address space to test the dtb
        try:
            if profile_cls.metadata("memory_model") == "64bit":
                virtual_as = addrspace.BaseAddressSpace.GetPlugin(
                    "AMD64PagedMemory")(
                    session=self.session,
                    base=self.session.physical_address_space,
                    dtb=dtb)
            elif profile_cls.metadata("pae"):
                virtual_as = addrspace.BaseAddressSpace.GetPlugin(
                    "IA32PagedMemoryPae")(
                    session=self.session,
                    base=self.session.physical_address_space,
                    dtb=dtb)
            else:
                virtual_as = addrspace.BaseAddressSpace.GetPlugin(
                    "IA32PagedMemory")(
                    session=self.session,
                    base=self.session.physical_address_space,
                    dtb=dtb)

        except addrspace.Error:
            return

        # Do some basic checks of the address space.
        if not self.verify_address_space(profile_cls, eprocess, virtual_as):
            return

        return virtual_as

    def update_session(self):
        """Try to update the session from the profile guess."""
        logging.info("Examining image for possible profiles.")

        # If we have a kdbg and a valid kernel address space we can try a faster
        # approach:
        if (self.session.GetParameter("kdbg") and
            self.session.GetParameter("dtb")):
            for profile, virtual_as, eprocess in self.guess_profile_from_kdbg(
                self.session.GetParameter("kdbg")):
                with self.session.state as state:
                    state.profile = profile

                self.session.kernel_address_space = virtual_as
                self.session.default_address_space = virtual_as
                logging.info("Autoselected profile %s", profile)

                return True

        for profile, virtual_as, eprocess in self.guess_profile():
            with self.session.state as state:
                state.profile = profile

            self.session.kernel_address_space = virtual_as
            self.session.default_address_space = virtual_as
            self.session.StoreParameter("dtb", int(virtual_as.dtb))

            # Try to set the correct _EPROCESS here.
            self.session.system_eprocess = self.session.profile._EPROCESS(
                vm=self.session.physical_address_space,
                offset=eprocess.obj_offset)

            logging.info("Autoselected profile %s", profile)

            return True

        logging.error("Can not autodetect profile - please set it "
                      "explicitely.")

    def render(self, renderer):
        if self.quick:
            renderer.format("Updating session profile and address spaces.\n")
            self.update_session()
            return

        renderer.table_header([("Potential Profile", "profile", "<30"),
                               ("Address Space", "as", "")])

        for profile, virtual_as, _ in self.guess_profile():
            renderer.table_row(profile.__class__.__name__,
                               virtual_as.name)


def generate_idle_lookup_list():
    """Generate the lookup table for _EPROCESS by loading all profiles."""
    eprocess_vtype = {}

    for name, cls in obj.Profile.classes.items():
        # Only care about the OS profiles.
        if cls.metadata("os") == "windows":
            kernel_profile = cls()
            eprocess = kernel_profile._EPROCESS(offset=0)

            list_entry = 'LIST_ENTRY32'
            if cls.metadata("memory_model") == "64bit":
                list_entry = 'LIST_ENTRY64'

            # For the idle method we want to know only the following members:
            eprocess_vtype[name] = [0, {
                    "ImageFileName": [eprocess.ImageFileName.obj_offset, [
                            'String', dict(length=16)]],
                    "ThreadListHead": [eprocess.ThreadListHead.obj_offset, [list_entry]],
                    "ActiveProcessLinks": [eprocess.ActiveProcessLinks.obj_offset, [list_entry]],
                    "UniqueProcessId": [eprocess.UniqueProcessId.obj_offset, ['unsigned int']],
                    "DirectoryTableBase": [
                        eprocess.Pcb.DirectoryTableBase.obj_offset, ["unsigned long"]],
                    }]

    return eprocess_vtype

# The following is generated from generate_idle_lookup_list() above.
eprocess_vtypes = {
  'VistaSP1x64': [0,
  {'ActiveProcessLinks': [232, ['LIST_ENTRY64']],
   'DirectoryTableBase': [40, ['unsigned long']],
   'ImageFileName': [568, ['String', {'length': 16}]],
   'ThreadListHead': [608, ['LIST_ENTRY64']],
   'UniqueProcessId': [224, ['unsigned int']]}],
 'VistaSP1x86': [0,
  {'ActiveProcessLinks': [160, ['LIST_ENTRY32']],
   'DirectoryTableBase': [24, ['unsigned long']],
   'ImageFileName': [332, ['String', {'length': 16}]],
   'ThreadListHead': [360, ['LIST_ENTRY32']],
   'UniqueProcessId': [156, ['unsigned int']]}],
 'VistaSP1x86PAE': [0,
  {'ActiveProcessLinks': [160, ['LIST_ENTRY32']],
   'DirectoryTableBase': [24, ['unsigned long']],
   'ImageFileName': [332, ['String', {'length': 16}]],
   'ThreadListHead': [360, ['LIST_ENTRY32']],
   'UniqueProcessId': [156, ['unsigned int']]}],
 'VistaSP2x64': [0,
  {'ActiveProcessLinks': [232, ['LIST_ENTRY64']],
   'DirectoryTableBase': [40, ['unsigned long']],
   'ImageFileName': [568, ['String', {'length': 16}]],
   'ThreadListHead': [608, ['LIST_ENTRY64']],
   'UniqueProcessId': [224, ['unsigned int']]}],
 'VistaSP2x86': [0,
  {'ActiveProcessLinks': [160, ['LIST_ENTRY32']],
   'DirectoryTableBase': [24, ['unsigned long']],
   'ImageFileName': [332, ['String', {'length': 16}]],
   'ThreadListHead': [360, ['LIST_ENTRY32']],
   'UniqueProcessId': [156, ['unsigned int']]}],
 'Win2008R2SP0x64': [0,
  {'ActiveProcessLinks': [392, ['LIST_ENTRY64']],
   'DirectoryTableBase': [40, ['unsigned long']],
   'ImageFileName': [736, ['String', {'length': 16}]],
   'ThreadListHead': [776, ['LIST_ENTRY64']],
   'UniqueProcessId': [384, ['unsigned int']]}],
 'Win2008R2SP1x64': [0,
  {'ActiveProcessLinks': [392, ['LIST_ENTRY64']],
   'DirectoryTableBase': [40, ['unsigned long']],
   'ImageFileName': [736, ['String', {'length': 16}]],
   'ThreadListHead': [776, ['LIST_ENTRY64']],
   'UniqueProcessId': [384, ['unsigned int']]}],
 'Win2008SP1x64': [0,
  {'ActiveProcessLinks': [232, ['LIST_ENTRY64']],
   'DirectoryTableBase': [40, ['unsigned long']],
   'ImageFileName': [568, ['String', {'length': 16}]],
   'ThreadListHead': [608, ['LIST_ENTRY64']],
   'UniqueProcessId': [224, ['unsigned int']]}],
 'Win2008SP1x86': [0,
  {'ActiveProcessLinks': [160, ['LIST_ENTRY32']],
   'DirectoryTableBase': [24, ['unsigned long']],
   'ImageFileName': [332, ['String', {'length': 16}]],
   'ThreadListHead': [360, ['LIST_ENTRY32']],
   'UniqueProcessId': [156, ['unsigned int']]}],
 'Win2008SP2x64': [0,
  {'ActiveProcessLinks': [232, ['LIST_ENTRY64']],
   'DirectoryTableBase': [40, ['unsigned long']],
   'ImageFileName': [568, ['String', {'length': 16}]],
   'ThreadListHead': [608, ['LIST_ENTRY64']],
   'UniqueProcessId': [224, ['unsigned int']]}],
 'Win2008SP2x86': [0,
  {'ActiveProcessLinks': [160, ['LIST_ENTRY32']],
   'DirectoryTableBase': [24, ['unsigned long']],
   'ImageFileName': [332, ['String', {'length': 16}]],
   'ThreadListHead': [360, ['LIST_ENTRY32']],
   'UniqueProcessId': [156, ['unsigned int']]}],
 'Win7SP0x64': [0,
  {'ActiveProcessLinks': [392, ['LIST_ENTRY64']],
   'DirectoryTableBase': [40, ['unsigned long']],
   'ImageFileName': [736, ['String', {'length': 16}]],
   'ThreadListHead': [776, ['LIST_ENTRY64']],
   'UniqueProcessId': [384, ['unsigned int']]}],
 'Win7SP0x86': [0,
  {'ActiveProcessLinks': [184, ['LIST_ENTRY32']],
   'DirectoryTableBase': [24, ['unsigned long']],
   'ImageFileName': [364, ['String', {'length': 16}]],
   'ThreadListHead': [392, ['LIST_ENTRY32']],
   'UniqueProcessId': [180, ['unsigned int']]}],
 'Win7SP1x64': [0,
  {'ActiveProcessLinks': [392, ['LIST_ENTRY64']],
   'DirectoryTableBase': [40, ['unsigned long']],
   'ImageFileName': [736, ['String', {'length': 16}]],
   'ThreadListHead': [776, ['LIST_ENTRY64']],
   'UniqueProcessId': [384, ['unsigned int']]}],
 'Win7SP1x86': [0,
  {'ActiveProcessLinks': [184, ['LIST_ENTRY32']],
   'DirectoryTableBase': [24, ['unsigned long']],
   'ImageFileName': [364, ['String', {'length': 16}]],
   'ThreadListHead': [392, ['LIST_ENTRY32']],
   'UniqueProcessId': [180, ['unsigned int']]}],
 'Win8SP0x64': [0,
  {'ActiveProcessLinks': [744, ['LIST_ENTRY64']],
   'DirectoryTableBase': [40, ['unsigned long']],
   'ImageFileName': [1080, ['String', {'length': 16}]],
   'ThreadListHead': [1136, ['LIST_ENTRY64']],
   'UniqueProcessId': [736, ['unsigned int']]}],
 'WinXPSP2x86': [0,
  {'ActiveProcessLinks': [136, ['LIST_ENTRY32']],
   'DirectoryTableBase': [24, ['unsigned long']],
   'ImageFileName': [372, ['String', {'length': 16}]],
   'ThreadListHead': [400, ['LIST_ENTRY32']],
   'UniqueProcessId': [132, ['unsigned int']]}],
 'WinXPSP3x86': [0,
  {'ActiveProcessLinks': [136, ['LIST_ENTRY32']],
   'DirectoryTableBase': [24, ['unsigned long']],
   'ImageFileName': [372, ['String', {'length': 16}]],
   'ThreadListHead': [400, ['LIST_ENTRY32']],
   'UniqueProcessId': [132, ['unsigned int']]}],
 'WinXPSP3x86PAE': [0,
  {'ActiveProcessLinks': [136, ['LIST_ENTRY32']],
   'DirectoryTableBase': [24, ['unsigned long']],
   'ImageFileName': [372, ['String', {'length': 16}]],
   'ThreadListHead': [400, ['LIST_ENTRY32']],
   'UniqueProcessId': [132, ['unsigned int']]}]}




# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
@author:       Michael Cohen
@license:      GNU General Public License 2.0 or later
@contact:      scudette@gmail.com
@organization: Google Inc.

This file encapsulates various virtual file system operations for supported
linux versions. The code is basically copied from the kernel sources of the
relevant versions.
"""
from rekall import utils
from rekall.plugins.overlays import basic


class FileName(object):
    """An object to represent a filename."""
    MAX_DEPTH = 15

    def __init__(self, components=None, start_dentry=None):
        self.start_dentry = start_dentry
        self.mount_point = "/"
        self.deleted = False
        if components is None:
            components = []
        self.path_components = components

    def PrependName(self, component):
        self.path_components.insert(0, utils.SmartUnicode(component))

    def FormatName(self, root_dentry):
        # For sockets we need more info.
        if len(self.path_components) >= self.MAX_DEPTH:
            return obj.NoneObject(
                u"Depth exceeded at %s" % "/".join(self.path_components))

        if self.mount_point == "socket:":
            return "{0}/{1}[{2}]".format(
                self.mount_point, root_dentry.d_name.name.deref(),
                self.start_dentry.d_inode.i_ino)

        elif self.mount_point == "pipe:":
            return "{0}[{1}]".format(
                self.mount_point, self.start_dentry.d_inode.i_ino)

        elif self.mount_point == "anon_inode:":
            return u"anon_inode:%s" % self.start_dentry.d_name.name.deref()

        # This is the normal condition for files.
        else:
            return self.__str__()

    def __str__(self):
        if self.deleted:
            deleted = " (deleted) "
        else:
            deleted = ""

        return u"%s%s%s" % (self.mount_point,
                            "/".join(self.path_components), deleted)


class Linux3VFS(object):
    """This is the implementation specific VFS operations.

    Most of the code here is a direct copy of the methods in linux/fs/dcache.c

    http://lxr.free-electrons.com/source/fs/mount.h?v=3.7#L53
    http://lxr.free-electrons.com/source/fs/dcache.c?v=3.7#L2576
    """


    def get_path(self, task, filp):
        """Resolve the dentry, vfsmount relative to this task's chroot.

        Returns:
          An absolute path to the global filesystem mount. (I.e. we do not
          truncate the path at the chroot point as the kernel does).
        """
        return self._prepend_path(filp.f_path, task.fs.root)

    def _real_mount(self, vfsmnt):
        """Return the mount container of the vfsmnt object."""
        return basic.container_of(vfsmnt, "mount", "mnt").reference()

    def _mnt_has_parent(self, mnt):
        return  mnt != mnt.mnt_parent

    def _prepend_path(self, path, root):
        """Return the path of a dentry.

        http://lxr.free-electrons.com/source/fs/dcache.c?v=3.7#L2576
        """
        # Ensure we can not get into an infinite loop here by limiting the
        # depth.
        depth = 0
        dentry = path.dentry
        vfsmnt = path.mnt
        mnt = self._real_mount(vfsmnt)

        result = FileName(start_dentry=dentry)

        # Check for deleted dentry.
        if dentry.d_flags.DCACHE_UNHASHED and not dentry.is_root:
            result.deleted = True

        while (dentry != root.dentry or vfsmnt != root.mnt):
            # Control the depth.
            depth += 1
            if depth >= result.MAX_DEPTH:
                break

            if dentry == vfsmnt.mnt_root or dentry.is_root:
                # Global root?
                if not self._mnt_has_parent(mnt):
                    break

                dentry = mnt.mnt_mountpoint
                mnt = mnt.mnt_parent
                vfsmnt = mnt.mnt.reference()
                continue

            parent = dentry.d_parent
            result.PrependName(dentry.d_name.name.deref())

            dentry = parent

        # When we get here dentry is a root dentry and mnt is the mount point it
        # is mounted on. There are some special mount points we want to
        # highlight.
        result.mount_point = mnt.mnt_mountpoint.d_name.name.deref()

        return result.FormatName(dentry)


class Linux26VFS(Linux3VFS):
    """This is the implementation specific VFS operations.

    Most of the code here is a direct copy of the methods in linux/fs/dcache.c

    http://lxr.free-electrons.com/source/fs/dcache.c?v=2.6.26#L1782
    """

    def get_path(self, task, filp):
        """Resolve the dentry, vfsmount relative to this task's chroot.

        Returns:
          An absolute path to the global filesystem mount. (I.e. we do not
          truncate the path at the chroot point as the kernel does).
        """
        root = task.fs.root.deref()
        # For very old kernels (<=2.6.24) the root member is a dentry not a
        # struct path, so we need to synthesize one.
        if root.obj_type == "dentry":
            root = task.obj_profile.path()
            root.dentry = task.fs.root
            root.mnt = task.fs.rootmnt

        return self.__d_path(filp.f_path, root)

    def d_unhashed(self, dentry):
        return dentry.d_flags.DCACHE_UNHASHED

    def __d_path(self, path, root):
        """A literal copy of the __d_path function from kernel 2.6.26."""
        depth = 0
        dentry = path.dentry
        vfsmnt = path.mnt

        result = FileName(start_dentry=dentry)

        if not dentry.is_root and self.d_unhashed(dentry):
            result.deleted = True

        # Limit the recursion here to avoid getting stuck.
        while depth < result.MAX_DEPTH:
            if (dentry == root.dentry and vfsmnt == root.mnt):
                break

            if (dentry == vfsmnt.mnt_root or dentry.is_root):
                if vfsmnt.mnt_parent == vfsmnt:
                    break

                dentry = vfsmnt.mnt_mountpoint
                vfsmnt = vfsmnt.mnt_parent
                continue

            parent = dentry.d_parent;
            result.PrependName(dentry.d_name.name.deref())

            dentry = parent;

        # When we get here dentry is a root dentry and mnt is the mount point it
        # is mounted on. There are some special mount points we want to
        # highlight.
        result.mount_point = utils.SmartUnicode(
            vfsmnt.mnt_mountpoint.d_name.name.deref())

        return result.FormatName(dentry)



# Rekall Memory Forensics
#
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
@author:       Andrew Case
@license:      GNU General Public License 2.0 or later
@contact:      atcuno@gmail.com
@organization: Digital Forensics Solutions
"""
import re

from rekall import args
from rekall import obj
from rekall import plugin
from rekall import utils

from rekall.plugins import core


class AbstractLinuxCommandPlugin(plugin.PhysicalASMixin,
                                 plugin.ProfileCommand):
    """A base class for all linux based plugins."""
    __abstract = True

    @classmethod
    def is_active(cls, config):
        """We are only active if the profile is linux."""
        return (getattr(config.profile, "_md_os", None) == 'linux' and
                plugin.Command.is_active(config))


class LinuxFindDTB(AbstractLinuxCommandPlugin):
    """A scanner for DTB values.

    For linux, the dtb values are taken directly from the symbol file. Linux has
    a direct mapping between the kernel virtual address space and the physical
    memory.  This is the difference between the virtual and physical addresses
    (aka PAGE_OFFSET). This is defined by the __va macro:

    #define __va(x) ((void *)((unsigned long) (x) + PAGE_OFFSET))

    This one plugin handles both 32 and 64 bits.
    """

    __name = "find_dtb"

    def dtb_hits(self):
        """Tries to locate the DTB."""
        if self.profile.metadata("memory_model") == "32bit":
            PAGE_OFFSET = (self.profile.get_constant("_text") -
                           self.profile.get_constant("phys_startup_32"))

            yield self.profile.get_constant("swapper_pg_dir") - PAGE_OFFSET, None
        else:
            PAGE_OFFSET = (self.profile.get_constant("_text") -
                           self.profile.get_constant("phys_startup_64"))

            yield self.profile.get_constant("init_level4_pgt") - PAGE_OFFSET, None

    def verify_address_space(self, address_space=None, **kwargs):
        # There is not really much we can do if the address space is wrong, so
        # we just keep going.
        return True

    def render(self, renderer):
        renderer.table_header([("DTB", "dtv", "[addrpad]"),
                               ("Valid", "valid", "")])

        for dtb, _ in self.dtb_hits():
            address_space = core.GetAddressSpaceImplementation(self.profile)(
                session=self.session, base=self.physical_address_space, dtb=dtb)

            renderer.table_row(
                dtb, self.verify_address_space(address_space=address_space))

class LinuxPlugin(plugin.KernelASMixin, AbstractLinuxCommandPlugin):
    """Plugin which requires the kernel Address space to be loaded."""
    __abstract = True


class LinProcessFilter(LinuxPlugin):
    """A class for filtering processes."""

    __abstract = True

    @classmethod
    def args(cls, parser):
        super(LinProcessFilter, cls).args(parser)
        parser.add_argument("--pid",
                            action=args.ArrayIntParser, nargs="+",
                            help="One or more pids of processes to select.")

        parser.add_argument("--proc_regex", default=None,
                            help="A regex to select a process by name.")

        parser.add_argument("--phys_task",
                            action=args.ArrayIntParser, nargs="+",
                            help="Physical addresses of task structs.")

        parser.add_argument("--task", action=args.ArrayIntParser, nargs="+",
                            help="Kernel addresses of task structs.")

        parser.add_argument("--task_head", action=args.IntParser,
                            help="Use this as the process head. If "
                            "specified we do not use kdbg.")


    def __init__(self, pid=None, proc_regex=None, phys_task=None, task=None,
                 task_head=None, **kwargs):
        """Filters processes by parameters.

        Args:
           phys_task_struct: One or more task structs or offsets defined in
              the physical AS.

           pids: A list of pids.
           pid: A single pid.
        """
        super(LinProcessFilter, self).__init__(**kwargs)

        if isinstance(phys_task, (int, long)):
            phys_task = [phys_task]
        elif phys_task is None:
            phys_task = []

        if isinstance(task, (int, long)):
            task = [task]
        elif isinstance(task, obj.Struct):
            task = [task.obj_offset]
        elif task is None:
            task = []

        self.phys_task = phys_task
        self.task = task

        pids = []
        if isinstance(pid, list):
            pids.extend(pid)

        elif isinstance(pid, (int, long)):
            pids.append(pid)

        if self.session.pid and not pid:
            pids.append(self.session.pid)

        self.pids = pids
        self.proc_regex_text = proc_regex
        if isinstance(proc_regex, basestring):
            proc_regex = re.compile(proc_regex, re.I)

        self.proc_regex = proc_regex

        # Without a specified task head, we use the init_task from the symbol
        # table.
        if task_head is None:
            task_head = self.profile.get_constant("init_task")

        self.task_head = task_head

        # Sometimes its important to know if any filtering is specified at all.
        self.filtering_requested = (self.pids or self.proc_regex or
                                    self.phys_task or self.task)

    def list_tasks(self):
        task = self.profile.task_struct(
            offset=self.task_head, vm=self.kernel_address_space)

        return iter(task.tasks)

    def filter_processes(self):
        """Filters task list using phys_task and pids lists."""
        # No filtering required:
        if not self.filtering_requested:
            for task in self.list_tasks():
                yield task
        else:
            # We need to filter by phys_task
            for offset in self.phys_task:
                yield self.virtual_process_from_physical_offset(offset)

            for offset in self.task:
                yield self.profile.task_struct(vm=self.kernel_address_space,
                                               offset=int(offset))

            # We need to filter by pids
            for task in self.list_tasks():
                if int(task.pid) in self.pids:
                    yield task
                elif self.proc_regex and self.proc_regex.match(
                    utils.SmartUnicode(task.comm)):
                    yield task


    def virtual_process_from_physical_offset(self, physical_offset):
        """Tries to return an task in virtual space from a physical offset.

        We do this by reflecting off the list elements.

        Args:
           physical_offset: The physcial offset of the process.

        Returns:
           an _TASK object or a NoneObject on failure.
        """
        physical_task = self.profile.eprocess(offset=int(physical_offset),
                                              vm=self.kernel_address_space.base)

        # We cast our list entry in the kernel AS by following Flink into the
        # kernel AS and then the Blink. Note the address space switch upon
        # dereferencing the pointer.
        our_list_entry = physical_task.tasks.next.dereference(
            vm=self.kernel_address_space).prev.dereference()

        # Now we get the task_struct object from the list entry.
        return our_list_entry.dereference_as("task_struct", "tasks")


class HeapScannerMixIn(object):
    """A mixin for converting a scanner into a heap only scanner."""
    def scan(self, **kwargs):
        for vma in self.task.mm.mmap.walk_list("vm_next"):
            # Only use the vmas inside the heap area.
            if (vma.vm_start >= self.task.mm.start_brk or
                vma.vm_end <= self.task.mm.brk):
                for hit in super(HeapScannerMixIn, self).scan(
                    offset=vma.vm_start, maxlen=vma.vm_end-vma.vm_start):
                    yield hit


class KernelAddressCheckerMixIn(object):
    """A plugin mixin which does kernel address checks."""

    def __init__(self, **kwargs):
        super(KernelAddressCheckerMixIn, self).__init__(**kwargs)

        # We use the module plugin to help us local addresses inside kernel
        # modules.
        self.module_plugin = self.session.plugins.lsmod(session=self.session)



# TODO: Deprecate this when all plugins have been converted.
class AbstractLinuxCommand(object):

    def __init__(self, *args, **kwargs):
        self.addr_space = utils.load_as(self._config)
        self.profile = self.addr_space.profile
        self.smap = self.profile.sys_map

    @classmethod
    def is_active(cls, session):
        """We are only active if the profile is Linux."""
        try:
            return session.profile and session.profile._md_os == 'linux'
        except profile.Error:
            return True


def offsetof(struct_name, list_member, profile):

    offset = profile.typeDict[struct_name][1][list_member][0]
    return offset

def bit_is_set(bmap, pos):

    mask = 1 << pos
    return bmap & mask

# returns a list of online cpus (the processor numbers)
def online_cpus(smap, addr_space):

    #later kernels..
    if "cpu_online_bits" in smap:
        bmap = obj.Object("unsigned long", offset=smap["cpu_online_bits"], vm=addr_space)

    elif "cpu_present_map" in smap:
        bmap = obj.Object("unsigned long",  offset=smap["cpu_present_map"], vm=addr_space)

    else:
        raise AttributeError, "Unable to determine number of online CPUs for memory capture"

    cpus = []
    for i in xrange(0, 8):
        if bit_is_set(bmap, i):
            cpus.append(i)

    return cpus

def walk_per_cpu_var(obj_ref, per_var, var_type):

    cpus = online_cpus(obj_ref.smap, obj_ref.addr_space)

    # get the highest numbered cpu
    max_cpu = cpus[-1]

    per_offsets = obj.Object(type_name='Array', targetType='unsigned long', count=max_cpu, offset=obj_ref.smap["__per_cpu_offset"], vm=obj_ref.addr_space)
    i = 0

    for i in cpus:

        offset = per_offsets[i]

        addr = obj_ref.smap["per_cpu__" + per_var] + offset.v()
        var = obj.Object(var_type, offset=addr, vm=obj_ref.addr_space)

        yield i, var


def S_ISDIR(mode):
    return (mode & linux_flags.S_IFMT) == linux_flags.S_IFDIR

def S_ISREG(mode):
    return (mode & linux_flags.S_IFMT) == linux_flags.S_IFREG








# Rekall Memory Forensics
# Copyright (C) 2007,2008 Volatile Systems
# Copyright (C) 2010,2011,2012 Michael Hale Ligh <michael.ligh@mnin.org>
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

from rekall import obj
from rekall import utils
from rekall import scan

from rekall.plugins.windows import common
from rekall.plugins.windows.gui import win32k_core
from rekall.plugins.windows.gui import windowstations


class PoolScanAtom(common.PoolScanner):
    """Pool scanner for atom tables"""

    def __init__(self, **kwargs):
        super(PoolScanAtom, self).__init__(**kwargs)

        self.checks = [
            ('PoolTagCheck', dict(tag=self.profile.get_constant(
                        "PoolTag_Atom"))),

            ('CheckPoolSize', dict(
                    min_size=self.profile.get_obj_size("_RTL_ATOM_TABLE"))),

            ('CheckPoolType', dict(paged=True, non_paged=True, free=True)),
            ]


class AtomScan(common.PoolScannerPlugin):
    """Pool scanner for _RTL_ATOM_TABLE"""

    allocation = ['_POOL_HEADER', '_RTL_ATOM_TABLE']

    __name = "atomscan"

    @classmethod
    def args(cls, parser):
        parser.add_argument("-S", "--sort-by",
                            choices=["atom", "refcount", "offset"],
                            default = "offset",
                            help = "Sort by [offset | atom | refcount]")

    def __init__(self, sort_by=None, **kwargs):
        super(AtomScan, self).__init__(**kwargs)
        self.sort_by = sort_by
        self.profile = win32k_core.Win32GUIProfile(self.profile)

    def generate_hits(self):
        scanner = PoolScanAtom(profile=self.profile, session=self.session,
                               address_space=self.address_space)

        for pool_header in scanner.scan():
            # Note: all OS after XP, there are an extra 8 bytes (for 32-bit)
            # or 16 bytes (for 64-bit) between the _POOL_HEADER and _RTL_ATOM_TABLE.
            # This is variable length structure, so we can't use the bottom-up
            # approach as we do with other object scanners - because the size of an
            # _RTL_ATOM_TABLE differs depending on the number of hash buckets.

            build = self.profile.metadatas('major', 'minor')
            fixup = 0

            if self.profile.metadata('memory_model') == '32bit':
                if build > (5, 1):
                    fixup = 8
            else:
                if build > (5, 1):
                    fixup = 16

            atom_table = self.profile._RTL_ATOM_TABLE(
                offset=pool_header.obj_offset + pool_header.size() + fixup,
                vm=pool_header.obj_vm)

            # There's no way to tell which session or window station
            # owns an atom table by *just* looking at the atom table,
            # so we have to instantiate it from the default kernel AS.
            if atom_table.is_valid():
                yield atom_table

    def render(self, renderer):

        renderer.table_header(
            [("TableOfs(P)", "physical_offset", "[addr]"),
             ("AtomOfs(V)", "virtual_offset", "[addrpad]"),
             ("Atom", "atom", "[addr]"),
             ("Refs", "refs", "6"),
             ("Pinned", "pinned", "6"),
             ("Name", "name", ""),
             ])

        for atom_table in self.generate_hits():
            # This defeats the purpose of having a generator, but
            # its required if we want to be able to sort. We also
            # filter string atoms here.
            atoms = []
            for atom in atom_table.atoms(vm=self.session.kernel_address_space):
                if atom.is_string_atom():
                    atoms.append(atom)

            if self.sort_by == "atom":
                attr = "Atom"
            elif self.sort_by == "refcount":
                attr = "ReferenceCount"
            else:
                attr = "obj_offset"

            for atom in sorted(atoms, key = lambda x: getattr(x, attr)):
                renderer.table_row(atom_table.obj_offset,
                                   atom.obj_offset,
                                   atom.Atom, atom.ReferenceCount,
                                   atom.Pinned,
                                   atom.Name)



class Atoms(windowstations.WndScan):
    """Print session and window station atom tables"""

    __name = "atoms"

    def find_atoms(self):
        seen = set()

        # Find the atom tables that belong to each window station
        for wndsta, session_space in self.generate_hits():
            # Get the atom table from the windows station object. Deduplicate by
            # physical address.
            table_physical_offset = session_space.vtop(wndsta.pGlobalAtomTable)
            if table_physical_offset in seen:
                continue

            seen.add(table_physical_offset)

            # The atom table is dereferenced in the proper
            # session space
            atom_table = wndsta.AtomTable(vm=session_space)

            if atom_table.is_valid():
                for atom in atom_table.atoms(vm=session_space):
                    ## Filter string atoms
                    if not atom.is_string_atom():
                        continue

                    yield table_physical_offset, atom, wndsta

        # Find atom tables not linked to specific window stations.
        # This finds win32k!UserAtomHandleTable.
        for table in self.session.plugins.atomscan().generate_hits():
            table_physical_offset = table.obj_vm.vtop(table.obj_offset)
            if table_physical_offset not in seen:
                for atom in table.atoms(vm=session_space):
                    ## Filter string atoms
                    if not atom.is_string_atom():
                        continue

                    yield (table_physical_offset, atom,
                           obj.NoneObject("No windowstation"))

    def render(self, renderer):
        renderer.table_header(
            [("Offset(P)", "physical_offset", "[addrpad]"),
             ("Session", "session", "<10"),
             ("WindowStation", "windows_station", "<18"),
             ("Atom", "atom", "[addr]"),
             ("RefCount", "ref_count", "<10"),
             ("HIndex", "hindex", "<10"),
             ("Pinned", "pinned", "<10"),
             ("Name", "name", ""),
             ])

        for table_physical_offset, atom, window_station in self.find_atoms():
            renderer.table_row(
                table_physical_offset,
                window_station.dwSessionId,
                window_station.Name.v(vm=atom.obj_vm),
                atom.Atom,
                atom.ReferenceCount,
                atom.HandleIndex,
                atom.Pinned,
                atom.Name.v(vm=atom.obj_vm))



# Rekall Memory Forensics
# Copyright (C) 2008 Volatile Systems
# Copyright (c) 2008 Brendan Dolan-Gavitt <bdolangavitt@wesleyan.edu>
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""
@author:       AAron Walters and Brendan Dolan-Gavitt
@license:      GNU General Public License 2.0 or later
@contact:      awalters@volatilesystems.com,bdolangavitt@wesleyan.edu
@organization: Volatile Systems
"""

from rekall.plugins.windows import common
from rekall.plugins.windows.registry import registry


class PoolScanHive(common.PoolScanner):
    checks = [ ('PoolTagCheck', dict(tag = "CM10")) ]

    def scan(self):
        for pool_obj in super(PoolScanHive, self).scan():
            # The _HHIVE is immediately after the pool header.
            hhive = self.profile._CMHIVE(
                offset=pool_obj.obj_offset + pool_obj.size(),
                vm=self.address_space)

            if hhive.Hive.Signature != 0xbee0bee0:
                continue

            yield hhive


class HiveScan(common.PoolScannerPlugin):
    """ Scan Physical memory for _CMHIVE objects (registry hives)

    You will need to obtain these offsets to feed into the hivelist command.
    """

    __name = "hivescan"

    meta_info = dict(
        author = 'Brendan Dolan-Gavitt',
        copyright = 'Copyright (c) 2007,2008 Brendan Dolan-Gavitt',
        contact = 'bdolangavitt@wesleyan.edu',
        license = 'GNU General Public License 2.0 or later',
        url = 'http://moyix.blogspot.com/',
        os = 'WIN_32_XP_SP2',
        version = '1.0',
        )

    def __init__(self, **kwargs):
        super(HiveScan, self).__init__(**kwargs)

        # Install our specific implementation of registry support.
        self.profile = registry.RekallRegisteryImplementation(self.profile)

    def generate_hits(self, address_space=None):
        """Yields potential _HHIVE objects."""
        # Scan for these in the kernel address space.
        address_space = address_space or self.physical_address_space
        scanner = PoolScanHive(profile=self.profile, address_space=address_space,
                               session=self.session)

        return scanner.scan()

    def list_hives(self, address_space=None):
        """Scans the address space for potential hives."""
        # Just return the hive list from the session.
        if self.session and self.session.hive_list is not None:
            for hive in self.session.hive_list:
                yield hive

            return

        hives = set()
        for phive in self.generate_hits(self.physical_address_space):
            # Reflect through the kernel address space to get the hive in kernel
            # address space.
            hive = phive.HiveList.reflect(vm=self.kernel_address_space).dereference_as(
                "_CMHIVE", "HiveList")

            if hive in hives:
                continue

            hives.add(hive)
            yield hive

        # Add the hives to the session for next time.
        self.session.hive_list = list(hives)

    def render(self, outfd):
        outfd.write("{0:10} {1:10} {2}\n".format("Offset(V)", "Offset(P)", "Name"))
        for phive in self.generate_hits():
            # Make the hive in the kernel address space by reflecting through
            # the HiveList. Note that this may not return to the original
            # object!
            hive = phive.HiveList.reflect(vm=self.kernel_address_space).dereference_as(
                "_CMHIVE", "HiveList")

            outfd.write("{0:#010x} {1:#010x} {2}\n".format(
                    hive.obj_offset, phive.obj_offset, hive.Name))



# Rekall Memory Forensics
# Copyright (c) 2010, 2011, 2012 Michael Ligh <michael.ligh@mnin.org>
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

from rekall import obj
from rekall import utils
from rekall.plugins.windows import common

#--------------------------------------------------------------------------------
# vtypes
#--------------------------------------------------------------------------------

# This type is defined in Win2K3SP0x86 and VistaSP2x86, but
# it applies to many other profiles in which it is not defined
# in the public PDBs.
timer_types = {
    '_KTIMER_TABLE_ENTRY' : [ 0x10, {
    'Entry' : [ 0x0, ['_LIST_ENTRY']],
    'Time'  : [ 0x8, ['_ULARGE_INTEGER']],
    }]}

#--------------------------------------------------------------------------------
# timers
#--------------------------------------------------------------------------------

class Timers(common.WindowsCommandPlugin):
    """Print kernel timers and associated module DPCs"""

    __name = "timers"

    def __init__(self, **kwargs):
        """Print kernel timers and associated module DPCs."""
        super(Timers, self).__init__(**kwargs)
        self.profile.add_types(timer_types)

    @classmethod
    def is_active(cls, config):
        """Timers are currently only supported on 32 bit windows."""
        profile = config.profile

        return (profile and profile.metadata("os") == 'windows' and
                profile.metadata("memory_model") == "32bit")

    def find_list_head(self, nt_mod, func, sig):
        """
        Find the KiTimerTableListHead given an exported
        function as a starting point and a small signature.

        @param nt_mod: _LDR_DATA_TABLE_ENTRY object for NT module
        @param func: function name exported by the NT module
        @param sig: byte string/pattern to use for finding the symbol
        """

        # Lookup the exported function
        func_rva = nt_mod.getprocaddress(func)
        if func_rva == None:
            return None

        func_addr = func_rva + nt_mod.DllBase

        # Read enough of the function prolog
        data = nt_mod.obj_vm.read(func_addr, 200)

        # Scan for the byte signature
        n = data.find(sig)
        if n == -1:
            return None

        return self.profile.Object('address', func_addr + n + len(sig), nt_mod.obj_vm)

    def calculate(self):
        # Get the OS version we're analyzing
        version = (self.profile.metadata('major', 0),
                   self.profile.metadata('minor', 0))

        modules = self.session.plugins.modules(session=self.session)
        mod_addrs = modules.addresses()
        modlist = list(modules.lsmod())

        # KTIMERs collected
        timers = []

        # Valid KTIMER.Header.Type values
        TimerNotificationObject    = 8
        TimerSynchronizationObject = 9
        valid_types = (TimerNotificationObject, TimerSynchronizationObject)

        if version == (5,1) or (version == (5,2) and
                                self.profile.metadata.get('major', 0) == 3789):
            """
            On XP SP0-SP3 x86 and Windows 2003 SP0, KiTimerTableListHead
            is an array of 256 _LIST_ENTRY for _KTIMERs.
            """
            KiTimerTableListHead = self.find_list_head(modlist[0],
                                        "KeUpdateSystemTime",
                                        "\x25\xFF\x00\x00\x00\x8D\x0C\xC5")

            lists = self.profile.Object("Array", offset = KiTimerTableListHead,
                                        vm = self.kernel_address_space,
                                        target = '_LIST_ENTRY',
                                        count = 256)

            for l in lists:
                for t in l.list_of_type("_KTIMER", "TimerListEntry"):
                    if t not in timers:
                        timers.append(t)

        elif version == (5,2) or version == (6,0):
            """
            On XP x64, Windows 2003 SP1-SP2, and Vista SP0-SP2, KiTimerTableListHead
            is an array of 512 _KTIMER_TABLE_ENTRY structs.
            """
            KiTimerTableListHead = self.find_list_head(modlist[0],
                                        "KeCancelTimer",
                                        "\xC1\xE7\x04\x81\xC7")

            lists = self.profile.Object("Array", offset = KiTimerTableListHead,
                                        vm = self.kernel_address_space,
                                        target = '_KTIMER_TABLE_ENTRY',
                                        count = 512)

            for l in lists:
                for t in l.Entry.list_of_type("_KTIMER", "TimerListEntry"):
                    timers.append(t)

        elif version == (6,1):
            """
            On Windows 7, there is no more KiTimerTableListHead. The list is
            at _KPCR.PrcbData.TimerTable.TimerEntries (credits to Matt Suiche
            for this one. See http://pastebin.com/FiRsGW3f).
            """
            kpcrscan = self.session.plugins.kpcrscan(session=self.session)
            for kpcr in kpcrscan.generate_hits():
                # Note that if this is not the correct kpcr, the NoneObject will
                # break out of the following loop:
                for table in kpcr.PrcbData.TimerTable.TimerEntries:
                    for t in table.Entry.list_of_type("_KTIMER", "TimerListEntry"):
                        timers.append(t)
        else:
            raise RuntimeError("Timers are not supported on this Profile.")

        # Sort the timers by address to make them easier to inspect.
        for timer in sorted(timers):
            # Sanity check on the timer type
            if timer.Header.Type not in valid_types:
                continue

            # Ignore timers without DPCs
            if not timer.Dpc.is_valid() or not timer.Dpc.DeferredRoutine.is_valid():
                continue

            # Lookup the module containing the DPC
            module = modules.find_module(timer.Dpc.DeferredRoutine)

            yield timer, module

    def render(self, renderer):
        renderer.table_header([("Offset", "offset", "[addrpad]"),
                               ("DueTime(H)          ", "due_high", "#020x"),
                               ("Period(ms)", "period", ">10"),
                               ("Signaled", "signaled", ">10"),
                               ("Routine", "routine", "[addrpad]"),
                               ("Module", "module", "20")])

        for timer, module in self.calculate():
            if timer.Header.SignalState.v():
                signaled = "Yes"
            else:
                signaled = "-"

            if module:
                module_name = module.BaseDllName
            else:
                module_name = "UNKNOWN"

            renderer.table_row(
                timer,
                timer.DueTime.QuadPart,
                timer.Period,
                signaled,
                timer.Dpc.DeferredRoutine,
                module_name)



#!/usr/bin/env python2

# Rekall Memory Forensics
# Copyright 2013 Google Inc. All Rights Reserved.
#
# Author: Michael Cohen scudette@google.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

__author__ = "Michael Cohen <scudette@google.com>"
import site

from rekall import plugin
from rekall.plugins import core

try:
    from rekall import ipython_support
except ImportError:
    ipython_support = None


class Notebook(plugin.Command):
    """Launch the IPython notebook."""

    __name = "notebook"

    @classmethod
    def is_active(cls, session):
        """Only active when the IPython notebook is actually installed."""
        try:
            import IPython.html.notebookapp  # pylint: disable=unused-variable

            return bool(ipython_support)
        except ImportError:
            return False


    def render(self, renderer):
        renderer.format("Starting IPython notebook.")
        renderer.format("Press Ctrl-c to return to the interactive shell.")
        ipython_support.NotebookSupport(self.session)


class Rekall(plugin.Command):
    """Starts a new rekall analysis session."""

    __name = "rekal"

    interactive = True

    @classmethod
    def args(cls, parser):
        super(Rekall, cls).args(parser)

        parser.add_argument("--filename",
                            help="The name of the image file to analyze.")

        parser.add_argument("--profile", default=None,
                            help="The name of the profile to load.")

        parser.add_argument("--pager", default=None,
                            help="The name of a program to page output "
                            "(e.g. notepad or less).")

    def __init__(self, filename=None, profile=None, verbose=False,
                 pager=None, **kwargs):
        super(Rekall, self).__init__(**kwargs)

        self.filename = filename
        self.profile = profile
        self.verbose = verbose
        self.pager = pager

    def render(self, renderer):
        renderer.format("Initializing Rekall session.\n")

        with self.session.state as state:
            state.logging = "DEBUG" if self.verbose else "WARN"
            state.pager = self.pager
            state.filename = self.filename

            # Clear the profile from the session.
            self.session.profile = None
            state.profile = self.profile

        renderer.format("Done!\n")

# This sets python's built in help command so we can use the help command in the
# shell.

# pylint: disable=protected-access

class RekallHelper(site._Helper):
    """A More useful default help function."""
    HELP_MESSAGE = """Welocome to Rekall Memory Forensics.

To get started:

- Initialize the Rekall session using the rekall plugin. e.g.:

Win7SP1x64:pmem 13:36:23> rekall filename=r"\\\\.\\pmem", profile="Win7SP1x64", pager="notepad"

- Select a plugin to run by tying it in. e.g.:

Win7SP1x64:pmem 13:39:26> plugins.pslist

- You can complete any command by tapping Tab twice. Useful completions include:
  - File names on disk.
  - Plugin names.
  - Plugin parameters.

- Adding a ? after any plugin will print help about this plugin.

- You can get help on any module or object by typing:

help object

Some interesting topics to get you started, explaining some rekall specific
concepts:

help addrspace - The address space.
help obj       - The rekall objects.
help profile   - What are Profiles?
"""

    def __call__(self, item=None, **kwargs):
        if item is None:
            print self.HELP_MESSAGE
        else:
            print core.Info(item=item)

site._Helper = RekallHelper
